"""
This submodule contains the Cython code equivalent to the original R code in InlineFunctions.R from the original MHN repo
as well as some functions to solve linear equations involving [I-Q].
"""
# author(s): Stefan Vocht

cimport cython
import numpy as np
cimport numpy as np

from scipy.linalg.cython_blas cimport dcopy, dscal, daxpy, ddot
from libc.stdlib cimport malloc, free
from libc.math cimport exp

np.import_array()

cdef void internal_kron_vec(double[:, :] theta_mat, int i, double[:] x_vec, double *pout, bint diag, bint transp):
    """
    This function multiplies the kronecker-product you get from the ith row of theta with a vector

    :param theta_mat: matrix containing the theta values
    :param i: row of theta used for the kronecker-product
    :param x_vec: vector that is multiplied with the kronecker-product matrix
    :param pout: vector that will contain the result of the multiplication
    :param diag: if False, the diagonal of the kronecker-product matrix is set to zero
    :param transp: if True, the kronecker-product matrix is transposed
    :return:
    """
    # inizialize some constants used in this function
    cdef double[:] theta_i = theta_mat[i, :]
    cdef int n = theta_mat.shape[0]
    cdef int nx = 1 << n
    cdef int nxhalf = nx / 2
    cdef double mOne = -1
    cdef double zero = 0

    cdef int incx = 1
    cdef int incx2 = 2
    cdef int j

    cdef double *ptmp = <double *> malloc(nx * sizeof(double))
    cdef double *px1
    cdef double *px2
    cdef double *shuffled_vec
    cdef double *old_vec
    cdef double *swap_vec
    cdef double theta

    # for the shuffle algorithm we have to initialize the pointers correctly
    if n & 1 == 1:
        swap_vec = ptmp
        shuffled_vec = &pout[0]
    else:
        swap_vec = &pout[0]
        shuffled_vec = ptmp

    old_vec = &x_vec[0]

    # use the shuffle algorithm to compute the product of the kronecker product with a vector
    for j in range(n):
        dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
        dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)

        theta = exp(theta_i[j])
        px1 = shuffled_vec
        px2 = shuffled_vec + nxhalf

        if j == i:
            if not transp:
                dcopy(&nxhalf, px1, &incx, px2, &incx)
                dscal(&nxhalf, &theta, px2, &incx)
                if diag:
                    dcopy(&nxhalf, px2, &incx, px1, &incx)
                    dscal(&nxhalf, &mOne, px1, &incx)
                else:
                    dscal(&nxhalf, &zero, px1, &incx)
            else:
                if diag:
                    theta *= -1
                    daxpy(&nxhalf, &mOne, px2, &incx, px1, &incx)
                    dscal(&nxhalf, &theta, px1, &incx)
                    dscal(&nxhalf, &zero, px2, &incx)
                else:
                    dcopy(&nxhalf,px2,&incx,px1,&incx)
                    dscal(&nxhalf,&theta,px1,&incx)
                    dscal(&nxhalf,&zero,px2,&incx)

        else:
            dscal(&nxhalf, &theta, px2, &incx)

        old_vec = shuffled_vec
        shuffled_vec = swap_vec
        swap_vec = old_vec

    free(ptmp)


def kron_vec(double[:, :] theta_mat, int i, double[:] x_vec, bint diag = False, bint transp = False) -> np.ndarray:
    """
    This function multiplies the kronecker-product you get from the ith row of theta with a vector.
    This is a Python wrapper for the more efficient Cython implementation.

    :param theta_mat: matrix containing the theta values
    :param i: row of theta used for the kronecker-product
    :param x_vec: vector that is multiplied with the kronecker-product matrix
    :param diag: if False, the diagonal of the kronecker-product matrix is set to zero
    :param transp: if True, the kronecker-product matrix is transposed
    :return: vector that will contain the result of the multiplication
    """
    cdef int n = theta_mat.shape[0]
    cdef np.ndarray[np.double_t] result = np.empty(2**n, dtype=np.double)
    internal_kron_vec(theta_mat, i, x_vec, &result[0], diag, transp)
    return result


cdef void loop_j(int i, int n, double *pr, double *pg):
    """
    This function is used in the gradient function (in Likelihood.pyx) to compute the gradient more efficiently

    :param i: current row of the gradient to be computed
    :param n: number of columns/rows of theta
    :param pr: a vector calculated in the gradient function
    :param pg: gradient matrix (output)
    :return:
    """   

    cdef int nx = 1 << n
    cdef int nxhalf = nx/2

    pg = pg + i * n

    cdef double one = 1.
    cdef double *ptmp = <double *> malloc(nx*sizeof(double))

    # instead of copying the shuffled vector back into the original
    # we change the following pointers in such a way that we do not have to copy the data back
    cdef double *shuffled_vec, *old_vec, *swap_vec
    cdef int incx = 1
    cdef int incx2 = 2
    cdef int incx0 = 0
    cdef int j

    old_vec = pr
    shuffled_vec = ptmp

    for j in range(n):
        # matrix shuffle
        dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
        dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)

        # sums as dot products with 1
        pg[j] = ddot(&nxhalf, shuffled_vec + nxhalf, &incx, &one, &incx0)
        if i == j:
            pg[j] += ddot(&nxhalf, shuffled_vec, &incx, &one, &incx0)

        swap_vec = old_vec
        old_vec = shuffled_vec
        shuffled_vec = swap_vec

    free(ptmp)


@cython.cdivision(True)
cdef void _compute_inverse(const double *theta, int n, const double *dg, const double *b, double *xout):
    """
    Internal function to compute the solution for [I-Q] x = b using forward substitution
    
    :param theta: thetas used to construct Q
    :param dg: vector containing the diagonal values of [I-Q]
    :param b: a double vector 
    :param xout: solution x as double vector
    """
    cdef int nx = 1 << n
    cdef int incx = 1

    cdef int i, j, k
    cdef int i_copy
    cdef int bit_setter
    cdef int modified_i

    cdef double theta_product
    cdef double xout_i

    # we need the exponential form of theta, so we compute it here once
    cdef double *exp_theta = <double *> malloc(n*n * sizeof(double))
    for i in range(n):
        for j in range(n):
            exp_theta[i*n + j] = exp(theta[i*n + j])

    # we compute the values of xout using forward substitution
    # at the beginning we initialize xout with the values of b
    # in each iteration i we compute the next final value of xout by dividing the current xout value by the ith diagonal entry
    # we then use that final value to update the values of all xout entries that represent states with exactly one more
    # mutation than the state represented by the ith xout entry
    # each index of those other xout entries can be obtained by flipping a bit in i from zero to one
    # if the jth bit is then set to one, this corresponds to the jth gene being mutated, so we have to add the product of
    # theta_jj with all theta_jk, where the kth gene is mutated in i, and with xout[i] to the temporary value of the
    # xout entry that we want to update
    dcopy(&nx, b, &incx, xout, &incx)
    for i in range(nx):
        # get the final value for xout by dividing with the diagonal entry
        xout[i] /= dg[i]
        # cache the value for later
        xout_i = xout[i]

        # update the other xout entries that differ by exactly one bit flipped to one
        # we use bit_setter to flip a zero at the jth position to become a one
        bit_setter = 1
        for j in range(n):
            modified_i = (i | bit_setter)
            if modified_i != i:
                # compute the product of all thetas which correspond to genes that are mutated in i
                theta_product = 1.
                i_copy = i
                for k in range(n):
                    if i_copy & 1:
                        theta_product *= exp_theta[j*n + k]
                    i_copy >>= 1
                xout[modified_i] += theta_product * exp_theta[j*n + j] * xout_i
            bit_setter <<= 1
    free(exp_theta)


@cython.cdivision(True)
cdef void _compute_inverse_t(const double *theta, int n, const double *dg, const double *b, double *xout):
    """
    Internal function to compute the solution for [I-Q]^T x = b using backward substitution

    :param theta: thetas used to construct Q
    :param dg: vector containing the diagonal values of [I-Q]
    :param b: a double vector 
    :param xout: solution x as double vector
    """
    cdef int nx = 1 << n
    cdef int incx = 1

    cdef int i, j, k
    cdef int i_copy
    cdef int bit_setter
    cdef int modified_i

    cdef double theta_product
    cdef double xout_i

    # we need the exponential form of theta, so we compute it here once
    cdef double *exp_theta = <double *> malloc(n * n * sizeof(double))
    for i in range(n):
        for j in range(n):
            exp_theta[i*n + j] = exp(theta[i*n + j])

    # initialize xout with the values of b
    dcopy(&nx, b, &incx, xout, &incx)

    # we compute the values of xout using backward substitution
    # at the beginning we initialize xout with the values of b
    # in each iteration i we compute the next final value of xout by dividing the current xout value by the ith diagonal entry
    # we then use that final value to update the values of all xout entries that represent states with exactly one less
    # mutation than the state represented by the ith xout entry
    # each index of those other xout entries can be obtained by flipping a bit in i from one to zero
    # if the jth bit is then set to zero, this corresponds to reversing the mutation of the jth gene, so we have to add the product of
    # theta_jj with all theta_jk, where the kth gene is mutated in i, and with xout[i] to the temporary value of the
    # xout entry that we want to update
    for i in range(nx-1, -1, -1):
        # get the final value for xout by dividing with the diagonal entry
        xout[i] /= dg[i]
        # cache the value for later
        xout_i = xout[i]
        # update the other xout entries that differ by exactly one bit flipped to zero
        # we use bit_setter to flip a one at the jth position to become a zero
        bit_setter = 1
        for j in range(n):
            modified_i = (i & (~bit_setter))
            if modified_i != i:
                # compute the product of all thetas which correspond to genes that are mutated in i
                theta_product = 1.
                i_copy = modified_i
                for k in range(n):
                    if i_copy & 1:
                        theta_product *= exp_theta[j*n + k]
                    i_copy >>= 1
                xout[modified_i] += theta_product * exp_theta[j*n + j] * xout_i
            bit_setter <<= 1
    free(exp_theta)


cpdef compute_inverse(double[:, :] theta, double[:] dg, double[:] b, double[:] xout, bint transp):
    """
    Computes the solution for [I - Q] x = b using forward (and backward) substitution.

    :param theta: thetas used to construct Q
    :param dg: vector containing the diagonal values of [I-Q]
    :param b: a double vector
    :param xout: double vector that will contain the solution after running this function
    :param transp: if True, returns solution for [I - Q]^T x = b
    """
    cdef int n = theta.shape[0]
    if transp:
        _compute_inverse_t(&theta[0, 0], n, &dg[0], &b[0], &xout[0])
    else:
        _compute_inverse(&theta[0, 0], n, &dg[0], &b[0], &xout[0])


"""
This submodule contains functions to learn an MHN using state-space restriction, including an optimization function
using the BFGS algorithm and L1 regularization functions
"""
# author(s): Stefan Vocht, Y. Linda Hu

import numpy as np
from scipy.optimize import minimize, OptimizeResult

from typing import Callable

from .state_containers import StateContainer, create_indep_model
from . import state_space_restriction


def L1(theta: np.ndarray, eps: float = 1e-05) -> float:
    """
    Computes the L1 penalty
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return np.sum(np.sqrt(theta_**2 + eps))


def L1_(theta: np.ndarray, eps: float = 1e-05) -> np.ndarray:
    """
    Derivative of the L1 penalty
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return theta_ / np.sqrt(theta_**2 + eps)


def reg_state_space_restriction_score(theta: np.ndarray, states: StateContainer, lam: float,
                                      n: int, score_grad_container: list) -> float:
    """
    Computes the score using state space restriction with L1 regularization

    :param theta: current theta
    :param states: states observed in the data
    :param lam: regularization parameter
    :param n: size of theta (nxn)
    :param score_grad_container: a list that enables this function to communicate with the gradient function
    :return: regularized score
    """
    theta = theta.reshape((n, n))

    # grad, score = state_space_restriction.gradient_and_score(theta, states)
    grad, score = state_space_restriction.gradient_and_score(theta, states)
    score_grad_container[0] = grad

    return -(score - lam * L1(theta))


def reg_state_space_restriction_gradient(theta: np.ndarray, states: StateContainer, lam: float,
                                         n: int, score_grad_container: list) -> np.ndarray:
    """
    Computes the gradient using state space restriction with L1 regularization

    :param theta: current theta
    :param states: states observed in the data
    :param lam: regularization parameter
    :param n: size of theta (nxn)
    :param score_grad_container: a list that enables this function to communicate with the score function
    :return: regularized gradient
    """

    n = n or int(np.sqrt(theta.size))
    theta_ = theta.reshape((n, n))

    grad = score_grad_container[0]
    if grad is None:
        grad, score = state_space_restriction.gradient_and_score(theta_, states)

    return -(grad - lam * L1_(theta_)).flatten()


def build_regularized_score_func(gradient_and_score_function: Callable):
    """
    This function gets a function which can compute a gradient and a score at the same time and returns a function
    which computes the score and adds a L1 regularization
    """
    def reg_score_func(theta: np.ndarray, states: StateContainer, lam: float,
                                          n: int, score_grad_container: list) -> float:
        """
        Computes the score using state space restriction with L1 regularization

        :param theta: current theta
        :param states: states observed in the data
        :param lam: regularization parameter
        :param n: size of theta (nxn)
        :param score_grad_container: a list that enables this function to communicate with the gradient function
        :return: regularized score
        """
        theta = theta.reshape((n, n))
        grad, score = gradient_and_score_function(theta, states)
        score_grad_container[0] = grad

        return -(score - lam * L1(theta))
    return reg_score_func


def build_regularized_gradient_func(gradient_and_score_function: Callable):
    """
    This function gets a function which can compute a gradient and a score at the same time and returns a function
    which computes the gradient and adds the gradient of the L1 regularization
    """
    def reg_gradient_func(theta: np.ndarray, states: StateContainer, lam: float,
                                             n: int, score_grad_container: list) -> np.ndarray:
        """
        Computes the gradient state space restriction with L1 regularization

        :param theta: current theta
        :param states: states observed in the data
        :param lam: regularization parameter
        :param n: size of theta (nxn)
        :param score_grad_container: a list that enables this function to communicate with the score function
        :return: regularized gradient
        """
        n = n or int(np.sqrt(theta.size))
        theta_ = theta.reshape((n, n))
        grad = score_grad_container[0]
        if grad is None:
            grad, score = gradient_and_score_function(theta_, states)
        return -(grad - lam * L1_(theta_)).flatten()
    return reg_gradient_func


def learn_MHN(states: StateContainer, init: np.ndarray = None, lam: float = 0, maxit: int = 5000,
              trace: bool = False, reltol: float = 1e-07, round_result: bool = True, callback: Callable = None,
              score_func: Callable = reg_state_space_restriction_score,
              jacobi: Callable = reg_state_space_restriction_gradient) -> OptimizeResult:
    """
    This function is used to train a MHN, it is very similar to the learn_MHN function from the original MHN

    :param states: a StateContainer object containing all mutation states observed in the data
    :param init: starting point for the training (initial theta)
    :param lam: tuning parameter lambda for regularization
    :param maxit: maximum number of training iterations
    :param trace: set to True to print convergence messages (see scipy.optimize.minimize)
    :param reltol: Gradient norm must be less than reltol before successful termination (see "gtol" scipy.optimize.minimize)
    :param round_result: if True, the result is rounded to two decimal places
    :param callback: function called after each iteration, must take theta as argument
    :param score_func: score function used for training
    :param jacobi: gradient function used for training
    :return: OptimizeResult object containing the trained model
    """

    n = states.get_data_shape()[1]

    if init is None:
        init = create_indep_model(states)

    # this container is given to the score and gradient function to communicate with each other
    score_and_gradient_container = [None, None]

    opt = minimize(fun=score_func, x0=init, args=(states, lam, n, score_and_gradient_container), method="L-BFGS-B",
                   jac=jacobi, options={'maxiter': maxit, 'disp': trace, 'gtol': reltol}, callback=callback)

    opt.x = opt.x.reshape((n, n))

    if round_result:
        opt.x = np.around(opt.x, decimals=2)

    return opt



"""
This submodule implements ModelConstruction.R from the original implementation in Cython.

It contains functions to generate random MHNs, build their transition rate matrix Q, the diagonal of Q, and to generate
and independence model for a given distribution
"""
# author(s): Stefan Vocht

cimport cython

from libc.math cimport exp

from scipy.sparse import diags as scipy_diags

import numpy as np
cimport numpy as np

np.import_array()

def random_theta(n: int, sparsity: float = 0, rounded: bool = True) -> np.ndarray:
    """
    Creates a random MHN with (log-transformed) parameters Theta

    :param n: size of Theta -> size of corresponding Q is 2^n
    :param sparsity: sparsity of Theta as percentage
    :param rounded: if True, the random Theta is rounded to two decimals
    :return: random Theta
    """
    theta = np.zeros((n, n))

    np.fill_diagonal(theta, -1)
    theta = theta.flatten()
    non_zeros = np.random.choice(np.argwhere(theta != -1).squeeze(), size=int((n**2 - n)*(1 - sparsity)))

    theta[non_zeros] = np.random.normal(size=non_zeros.size)
    theta = theta.reshape((n, n))
    np.fill_diagonal(theta, np.random.normal(size=n))

    if rounded:
        theta = np.around(theta, decimals=2)

    return theta


cpdef np.ndarray[np.double_t, ndim=1] q_subdiag(double[:, :] theta, int i):
    """
    Creates a single subdiagonal of Q from the ith row in Theta

    :return: subdiagonal of Q corresponding to the ith row of Theta
    """
    cdef double[:] row = theta[i, :]
    cdef int n = theta.shape[0]
    cdef int j

    # s is the subdiagonal of Q, the entries are calculated as described in eq. 2
    cdef np.ndarray[np.double_t, ndim=1] s = np.empty(2**n, dtype=np.double)
    s[0] = exp(row[i])

    for j in range(n):
        s[2**j: 2**(j+1)] = s[:2**j] * exp(row[j]) * (i != j)

    return s


def build_q(theta: np.ndarray) -> np.ndarray:
    """
    Build the transition rate matrix Q from its subdiagonals for a given Theta

    :param theta: matrix representing the MHN
    :return: rate matrix Q
    """
    n = theta.shape[0]

    subdiags = np.array(list(map(lambda i: q_subdiag(theta, i), range(n))))
    diag = -np.sum(subdiags, axis=0)

    # scipy's diag cuts the end of the subdiags that dont fit into the final matrix Q
    q = scipy_diags(np.vstack((diag, subdiags)), offsets=[0] + [-2**i for i in range(n)], shape=(2**n, 2**n))

    return q.toarray()


cpdef np.ndarray[np.double_t, ndim=1]  q_diag(double[:, :] theta):
    """
    get the diagonal of Q

    :param theta: theta representing the MHN
    """
    cdef int n = theta.shape[0]
    cdef int i
    cdef np.ndarray[np.double_t, ndim=1] dg = np.zeros((2**n), dtype=np.double)

    for i in range(n):
        # the diagonal elements are the negative sums of their columns
        dg = dg - q_subdiag(theta, i)

    return dg


def learn_indep(pD: np.ndarray) -> np.ndarray:
    """
    Learns an independence model from the data distribution, which assumes that no events interact.
    Used to initialize the parameters of the actual model before optimization

    :param pD: probability distribution of the events in the data
    :return: independence model
    """
    cdef int n = int(np.log2(pD.size))
    cdef int i
    cdef np.ndarray[np.double_t, ndim=2] theta = np.zeros((n, n), dtype=np.double)

    # for each event i, sum up the probabilities of events where gene i was mutated to get the total probability
    # theta_ii is the log-odd of gene i being mutated
    for i in range(n):
        pD = pD.T.flatten()
        pD = pD.reshape((2**(n-1), 2))

        perc = np.sum(pD[:, 1])

        # fill theta_ii with the log-odds of gene i being mutated in a sample
        theta[i, i] = np.log(perc / (1 - perc))

    return np.around(theta, decimals=2)


"""
This submodule contains Optimizer classes to learn an MHN from mutation data.
"""
# author(s): Stefan Vocht, Y. Linda Hu

from __future__ import annotations

import warnings
from enum import Enum
import abc

import numpy as np
import pandas as pd

from . import model

from mhn.ssr import regularized_optimization as reg_optim
from .ssr.state_containers import StateContainer, StateAgeContainer
from .ssr.state_space_restriction import CUDAError, cuda_available, CUDA_AVAILABLE
from mhn.ssr import state_space_restriction as marginalized_funcs

from mhn.ssr import matrix_exponential as mat_exp


class _Optimizer(abc.ABC):
    """
    This abstract Optimizer class is the base class for the other Optimizer classes and cannot be instantiated alone.
    """
    def __init__(self):
        self._data = None
        self._bin_datamatrix = None
        self.__result = None
        self._events = None

        self.__init_theta = None
        self.__custom_callback = None

        self.__backup_steps = -1
        self.__backup_filename = None
        self.__backup_always_new_file = False
        self.__backup_current_step = None

        self._gradient_and_score_func = None

    def set_init_theta(self, init: np.ndarray):
        """
        Use this method to set a theta as starting point for learning a new MHN. The theta must be in logarithmic form.

        If none is given, the optimization starts with an independence model, where the baseline hazard Theta_ii
        of each event is set to its empirical odds and the hazard ratios (off-diagonal entries) are set to exactly 1.
        """
        self.__init_theta = init
        return self

    def get_data_properties(self):
        """
        You can use this method to get some information about the loaded mutation data, e.g. how many events and samples
        are present in the data, how many mutations a sample has on average etc.

        :returns: a dictionary containing information about the data
        """
        if self._bin_datamatrix is None:
            return {}

        total_mutations_per_sample = np.sum(self._bin_datamatrix, axis=1)
        return {
            'samples': self._bin_datamatrix.shape[0],
            'events': self._bin_datamatrix.shape[1],
            'mutations per sample': {
                'mean': np.mean(total_mutations_per_sample),
                'median': np.median(total_mutations_per_sample),
                'max': np.max(total_mutations_per_sample),
                'min': np.min(total_mutations_per_sample)
            }
        }

    def set_callback_func(self, callback=None):
        """
        Use this method to set a callback function called after each iteration in the BFGS algorithm.
        The given function must take exactly one argument, namely the theta matrix computed in the last iteration.
        """
        if callback is not None and not hasattr(callback, '__call__'):
            raise ValueError("callback has to be a function!")
        self.__custom_callback = callback
        return self

    def save_progress(self, steps: int = -1, always_new_file: bool = False, filename: str = 'theta_backup.npy'):
        """
        If you want to regularly save the progress during training, you can use this function and define the number
        of steps between each progress save.

        :param filename: file name of the backup file
        :param steps: number of training iterations between each progress backup
        :param always_new_file: if this is True, every backup is stored in a separate file, else the file is overwritten each time
        :return: this optimizer object
        """
        self.__backup_steps = steps
        self.__backup_always_new_file = always_new_file
        self.__backup_filename = filename
        return self

    def __total_callback_func(self, theta: np.ndarray):
        if self.__custom_callback is not None:
            self.__custom_callback(theta)

        if self.__backup_steps > 0:
            self.__create_backup(theta)

    def __create_backup(self, theta: np.ndarray):
        self.__backup_current_step += 1
        if (self.__backup_current_step % self.__backup_steps) == 0:
            filename = self.__backup_filename
            if self.__backup_always_new_file:
                try:
                    idx = filename.index(".")
                    filename = filename[:idx] + f"_{self.__backup_current_step}" + filename[idx:]
                except ValueError:  # str.index raises ValueError if no "." is present in the filename
                    filename += f"_{self.__backup_current_step}.npy"
            with open(filename, 'wb') as f:
                np.save(f, theta)

    def train(self, lam: float = None, maxit: int = 5000, trace: bool = False,
              reltol: float = 1e-7, round_result: bool = True) -> model.MHN:
        """
        Use this function to learn a new MHN from the data given to this optimizer.

        :param lam: tuning parameter lambda for regularization (default: 1/(number of samples in the dataset))
        :param maxit: maximum number of training iterations
        :param trace: set to True to print convergence messages (see scipy.optimize.minimize)
        :param reltol: Gradient norm must be less than reltol before successful termination (see "gtol" scipy.optimize.minimize)
        :param round_result: if True, the result is rounded to two decimal places
        :return: trained model
        """
        if self._data is None:
            raise ValueError("You have to load data before training!")

        if lam is None:
            lam = 1 / self._data.get_data_shape()[0]

        self.__result = None
        self.__backup_current_step = 0

        if self.__custom_callback is None and self.__backup_steps < 1:
            callback_func = None
        else:
            callback_func = self.__total_callback_func

        score_func = reg_optim.build_regularized_score_func(self._gradient_and_score_func)
        gradient_func = reg_optim.build_regularized_gradient_func(self._gradient_and_score_func)

        result = reg_optim.learn_MHN(self._data, self.__init_theta, lam, maxit, trace, reltol,
                                     round_result, callback_func, score_func, gradient_func)

        self.__backup_current_step = None

        self.__result = model.MHN(
            log_theta=result.x,
            events=self._events,
            meta={
                "lambda": lam,
                "init": self.__init_theta,
                "maxit": maxit,
                "reltol": reltol,
                "score": result.fun,
                "message": result.message,
                "status": result.status,
                "nit": result.nit
            })

        return self.__result

    @property
    def result(self) -> model.MHN:
        """
        The resulting MHN after training, same as the return value of the train() method.
        This property mainly exists as a kind of backup to ensure that the result of the training is not lost, if the
        user forgets to save the returned value of the train() method in a variable.
        """
        return self.__result

    @property
    @abc.abstractmethod
    def training_data(self):
        """
        This property returns all the data given to this optimizer to train a new MHN.
        """
        pass

    @staticmethod
    def _preprocess_binary_matrix(data_matrix: np.ndarray) -> np.ndarray:
        """
        This function is used to make sure that the given data matrix is in the correct format.
        Correct format:
            a 2D numpy array with dtype=np.int32, which only contains 0s and 1s

        While the dtype will be changed automatically, if not np.int32, a matrix not being 2D or containing other values
        than 0s and 1s will raise a ValueError.

        :return: the given data_matrix with its dtype set to np.int32
        """
        if len(data_matrix.shape) != 2:
            raise ValueError("The given data matrix must be two-dimensional")
        # StateContainer only accepts numpy arrays with dtype=np.int32
        if data_matrix.dtype != np.int32:
            data_matrix = data_matrix.astype(dtype=np.int32)
            warnings.warn("The dtype of the given data matrix is changed to np.int32")
        if not set(data_matrix.flatten()).issubset({0, 1}):
            raise ValueError("The data matrix must only contain 0s and 1s")

        return data_matrix

    @abc.abstractmethod
    def set_device(self, device: "_Optimizer.Device"):
        """
        Set the device that should be used for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.
        """
        if not isinstance(device, _Optimizer.Device):
            raise ValueError(f"The given device is not an instance of {_Optimizer.Device}")

        return self

    class Device(Enum):
        """
        A small Enum which can represent device types.
        """
        AUTO, CPU, GPU = range(3)


class StateSpaceOptimizer(_Optimizer):
    """
    This optimizer uses state-space restriction to optimize an MHN on data with no age information for the individual
    samples.
    """
    def __init__(self):
        super().__init__()
        self._gradient_and_score_func = marginalized_funcs.gradient_and_score

    def load_data_matrix(self, data_matrix: np.ndarray | pd.DataFrame):
        """
        Load mutation data stored in a numpy array or pandas DataFrame, where the rows represent samples and
        columns represent genes.
        Mutations of genes are represented by 1s, intact genes are represented by 0s.

        :param data_matrix: either a pandas DataFrame or a two-dimensional numpy array which should have dtype=np.int32
        :return: this optimizer object
        """
        if isinstance(data_matrix, pd.DataFrame):
            self._events = data_matrix.columns.to_list()
            data_matrix = np.array(data_matrix, dtype=np.int32)
        data_matrix = self._preprocess_binary_matrix(data_matrix)
        self._data = StateContainer(data_matrix)
        self._bin_datamatrix = data_matrix
        return self

    def load_data_from_csv(self, src: str, delimiter: str = ',', **kwargs):
        """
        Load mutation data from a CSV file. The rows have to represent samples and the columns represent genes.
        Mutations of genes are represented by 1s, intact genes are represented by 0s.

        :param src: path to the CSV file
        :param delimiter:  delimiter used in the CSV file (default: ',')
        :param kwargs: all additional keyword arguments are passed on to pandas' read_csv() function
        :return: this optimizer object
        """
        df = pd.read_csv(src, delimiter=delimiter, **kwargs)
        self.load_data_matrix(df)
        return self

    def set_device(self, device: "StateSpaceOptimizer.Device"):
        """
        Set the device that should be used for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.
        """
        super().set_device(device)
        if device == _Optimizer.Device.GPU:
            if cuda_available() != CUDA_AVAILABLE:
                raise CUDAError(cuda_available())

            self._gradient_and_score_func = marginalized_funcs.cuda_gradient_and_score
        else:
            self._gradient_and_score_func = {
                _Optimizer.Device.AUTO: marginalized_funcs.gradient_and_score,
                _Optimizer.Device.CPU: marginalized_funcs.cython_gradient_and_score
            }[device]
        return self

    @property
    def training_data(self) -> np.ndarray:
        """
        This property returns all the data given to this optimizer to train a new MHN.
        """
        return self._bin_datamatrix


class DUAOptimizer(_Optimizer):
    """
    This optimizer can make use of age information for each state in the data and can learn an MHN from that using
    methods described in Rupp et al.(2021).
    """

    def __init__(self):
        super().__init__()
        # the matrix exponential functions expect an additional parameter eps for the accuracy of the result
        # therefore, we have two separate "gradient_and_score" members, one that takes eps as parameter, and one that
        # does not and which we use in the train() method
        # the eps can be specified as parameter in the train() method
        self.__gradient_and_score_func_with_eps = mat_exp.cython_gradient_and_score
        self._gradient_and_score_func = lambda theta, data: self.__gradient_and_score_func_with_eps(theta, data, 1e-2)
        self._state_ages = None

    def load_data(self, data_matrix: np.ndarray, state_ages: np.ndarray):
        """
        Load training data consisting of a data matrix containing the observed states (rows represent samples
        and columns genes) and an array containing the ages of each state in the data matrix. Thus, the number of ages
        must align with the number of rows in the given data matrix.

        :param data_matrix: two-dimensional numpy array which should have dtype=np.int32
        :param state_ages: one-dimensional numpy array which should have dtype=np.double
        """
        data_matrix = self._preprocess_binary_matrix(data_matrix)
        if data_matrix.shape[0] != state_ages.shape[0]:
            raise ValueError("The number of samples in the data matrix must align with the number of ages given")

        if not np.all(state_ages[:-1] <= state_ages[1:]):
            warnings.warn("The data is not sorted by age yet. This should be no problem, but if you expected your data"
                          "to be sorted by age, something went wrong.")
        self._data = StateAgeContainer(data_matrix, state_ages)
        self._bin_datamatrix = data_matrix
        self._state_ages = state_ages
        return self

    def set_device(self, device: _Optimizer.Device):
        """
        Set the device that should be used for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.
        """
        super().set_device(device)
        if device == _Optimizer.Device.GPU:
            raise NotImplementedError("There is currently no GPU version available for the DUA algorithm,"
                                      " might be added later")
        else:
            self.__gradient_and_score_func_with_eps = mat_exp.cython_gradient_and_score
        return self

    def train(self, lam: float = 0, eps: float = 1e-2, maxit: int = 5000, trace: bool = False,
              reltol: float = 1e-7, round_result: bool = True) -> model.MHN:
        """
        Use this function to learn a new MHN from the data given to this optimizer.

        :param lam: tuning parameter lambda for the L1 regularization
        :param eps: accuracy of the matrix exponential
        :param maxit: maximum number of training iterations
        :param trace: set to True to print convergence messages (see scipy.optimize.minimize)
        :param reltol: Gradient norm must be less than reltol before successful termination (see "gtol" scipy.optimize.minimize)
        :param round_result: if True, the result is rounded to two decimal places
        :return: trained model
        """
        self._gradient_and_score_func = lambda theta, data: self.__gradient_and_score_func_with_eps(theta, data, eps)
        return super().train(lam, maxit, trace, reltol, round_result)

    @property
    def training_data(self) -> (np.ndarray, np.ndarray):
        """
        This property returns all the data given to this optimizer to train a new MHN.
        """
        return self._bin_datamatrix, self._state_ages



# by Stefan Vocht
#
# this file acts as a header file for PerformanceCriticalCode.pyx
#


cdef void internal_kron_vec(double[:, :] theta_mat, int i, double[:] x_vec, double *pout, bint diag, bint transp)

cdef void loop_j(int i, int n, double *pr, double *pg)

cpdef compute_inverse(double[:, :] theta, double[:] dg, double[:] b, double[:] xout, bint transp)

cdef void _compute_inverse(const double *theta, int n, const double *dg, const double *b, double *xout)

cdef void _compute_inverse_t(const double *theta, int n, const double *dg, const double *b, double *xout)



"""
This part of the mhn package contains functions to compute and work with MHNs on the full state-space.

The structure of this part of the package is very similar to the original R implementation by Schill et al. (2019):

Likelihood:
    contains functions to compute the log-likelihood score and its gradient without state-space restriction as
    well as functions for matrix-vector multiplications with the transition rate matrix and [I-Q]^(-1)

ModelConstruction:
    contains functions to generate random MHNs, build their transition rate matrix Q, the diagonal of Q, and to generate
    and independence model for a given distribution

RegularizedOptimization:
    contains functions to learn an MHN for a given data distribution, implements the L1 regularization

UtilityFunctions:
    contains functions useful for preprocessing data to be used as training data

PerformanceCriticalCode:
    contains functions that dominate the runtime of score and gradient computations and must therefore be implemented
    efficiently
"""
# author(s): Stefan Vocht


from . import PerformanceCriticalCode
from . import Likelihood
from . import ModelConstruction
from . import RegularizedOptimization
from . import UtilityFunctions


"""
This file contains unit tests for the full state-space functions of the mhn package.
"""
# author: Stefan Vocht

import unittest
import numpy as np

from mhn.ssr import state_space_restriction
from mhn.original import Likelihood, ModelConstruction, PerformanceCriticalCode


class TestCudaGradient(unittest.TestCase):
    """
    Tests for the function cuda_gradient_and_score
    """
    def setUp(self) -> None:
        """
        Preparation for each test
        """
        np.random.seed(0)  # set random seed for reproducibility
        if state_space_restriction.cuda_available() != state_space_restriction.CUDA_AVAILABLE:
            self.skipTest("CUDA not available for testing")

    def test_compare_with_cython(self):
        """
        Compare the full state-space score and gradient of the CUDA implementation with those of the Cython implementation
        """
        n = 4
        theta = ModelConstruction.random_theta(n)
        pD = np.random.random(2**n)
        pD /= pD.sum()
        gradient1 = Likelihood.grad(theta, pD)
        score1 = Likelihood.score(theta, pD)
        gradient2, score2 = Likelihood.cuda_gradient_and_score(theta, pD)
        self.assertEqual(round(score1, 8), round(score2, 8))
        print(score1, score2)
        print(gradient2)
        np.testing.assert_array_equal(np.around(gradient1, decimals=8), np.around(gradient2, decimals=8))

    def test_forward_substitution(self):
        """
        Test the computation of [I-Q]^(-1) b of the CUDA implementation with the Cython implementation
        """
        n = 3
        theta = ModelConstruction.random_theta(n)
        pD = np.random.random(2**n)
        pD /= pD.sum()

        res1 = Likelihood.cuda_compute_inverse(theta, pD)
        dg = 1 - ModelConstruction.q_diag(theta)
        res2 = np.empty(2**n)
        PerformanceCriticalCode.compute_inverse(theta, dg, pD, res2, False)

        print(res1)
        print(res2)



# by Stefan Vocht
#
# this is the header file for state_space_restriction.pyx

from mhn.ssr.state_containers cimport State

cdef int get_mutation_num(State *state)

cdef void restricted_q_vec(double[:, :] theta, double[:] x, State *state, double *yout, bint diag= *, bint transp = *)

cdef void restricted_q_diag(double[:, :] theta, State *state, double *dg)


"""
This submodule contains functions and classes to store and convert mutation states used for state-space restriction.
It also contains a function to compute an independence model that can be used as a starting point for training
a new MHN.
"""
# author(s): Stefan Vocht

from libc.stdlib cimport malloc, free
from libc.string cimport memcpy

from mhn.ssr.state_containers cimport State

import numpy as np
import warnings

# STATE_SIZE is defined in setup.py, 
# the maximum number n of genes the MHN can handle is 32 * STATE_SIZE


cdef int compute_max_mutation_number(int[:, :] mutation_data):
    """
    This function is used to compute the maximum number of mutations in a single sample of the data.
    """
    cdef int max_mutation_num = 0
    cdef int local_sum
    cdef int i, j
    for i in range(mutation_data.shape[0]):
        local_sum = 0
        for j in range(mutation_data.shape[1]):
            local_sum += mutation_data[i, j]

        if local_sum > max_mutation_num:
            max_mutation_num = local_sum

    return max_mutation_num


cdef void fill_states(State *states, int[:, :] mutation_data):
    """
    This function fills the given (yet empty) states with the information from mutation_data.
    """
    cdef int i, j
    cdef State *current_state
    cdef int gene_num = mutation_data.shape[1]

    for i in range(mutation_data.shape[0]):
        current_state = states + i

        for j in range(STATE_SIZE):
            current_state[0].parts[j] = 0

        for j in range(gene_num):
            if mutation_data[i, j]:
                current_state[0].parts[j >> 5] |=  1 << (j & 31)


cdef void sort_by_age(State *states, double *ages, int state_num):
    """
    Simplistic sort algorithm to sort both states and ages according to the age values.
    """
    cdef int i, j
    cdef double tmp_age
    cdef State tmp_state
    cdef bint changing = True

    for j in range(state_num):
        changing = False
        for i in range(state_num-j-1):
            if ages[i] > ages[i+1]:
                changing = True
                tmp_age = ages[i]
                tmp_state = states[i]
                ages[i] = ages[i+1]
                states[i] = states[i+1]
                ages[i+1] = tmp_age
                states[i+1] = tmp_state
        if not changing:
            break


cdef class StateContainer:
    """
    This class is used as a wrapper such that the C array containing the States can be referenced in a Python script.

    It also makes sure that there aren't more than 32 mutations present in a single sample as this would break the algorithms
    """

    def __init__(self, int[:, :] mutation_data):
        """
        :param mutation_data: a 2d numpy array with dtype=np.int32 that contains only 0s and 1s, where rows represent samples, and columns represent events
        """
        # the number of columns (number of genes) must not exceed 32 * STATE_SIZE
        if mutation_data.shape[1] > (32 * STATE_SIZE):
            raise ValueError(f"The number of genes present in the mutation data must not exceed {32 * STATE_SIZE}")

        self.data_size = mutation_data.shape[0]
        self.gene_num = mutation_data.shape[1]

        self.max_mutation_num = compute_max_mutation_number(mutation_data)
        if self.max_mutation_num == 0:
            warnings.warn("Your data does not contain any mutations, something went probably wrong")
        elif self.max_mutation_num > 32:
            raise ValueError("A single sample must not contain more than 32 mutations")

        self.states = <State *> malloc(self.data_size * sizeof(State))

        if not self.states:
            raise MemoryError()

        fill_states(self.states, mutation_data)


    def get_data_shape(self):
        """
        returns a tuple containing the number of tumor samples and the number of genes stored in this object
        """
        return (self.data_size, self.gene_num)

    def get_max_mutation_num(self):
        """
        returns the maximum number of mutations present in a single sample, might be useful as a sanity check
        """
        return self.max_mutation_num

    def __dealloc__(self):
        free(self.states)


cdef class StateAgeContainer(StateContainer):
    """
    This class is used as a wrapper like the StateContainer class, but also contains age information for each sample.
    """

    def __init__(self, int[:, :] mutation_data, double[:] ages):
        """
        :param mutation_data: a 2d numpy array with dtype=np.int32 that contains only 0s and 1s, where rows represent samples, and columns represent events
        :param ages: a 1d numpy array with dtype=np.double that contains the ages of the samples present in mutation_data
        """
        super().__init__(mutation_data)
        if ages.shape[0] != self.data_size:
            raise ValueError("The number of given ages must align with the number of samples in the mutation data")
        self.state_ages = <double *> malloc(self.data_size * sizeof(double))
        if not self.state_ages:
            raise MemoryError()
        memcpy(self.state_ages, &ages[0], self.data_size * sizeof(double))
        sort_by_age(self.states, self.state_ages, ages.shape[0])

    def __dealloc__(self):
        free(self.state_ages)


def create_indep_model(StateContainer state_container):
    """
    Compute an independence model from the data stored in the StateContainer object, where the baseline hazard Theta_ii
    of each event is set to its empirical odds and the hazard ratios (off-diagonal entries) are set to exactly 1.
    The independence model is returned in logarithmic representation.

    :param state_container: StateContainer object containing the data on which the independence model is based
    :returns: an independence model in logarithmic representation
    """

    cdef int n = state_container.gene_num

    cdef int i, j
    cdef int sum_of_occurance

    theta = np.zeros((n, n))

    for i in range(n):
        sum_of_occurance = 0
        for j in range(state_container.data_size):
            sum_of_occurance += (state_container.states[j].parts[i >> 5] >> (i & 31)) & 1

        if sum_of_occurance == 0:
            warnings.warn(f"During independence model creation: event {i} never occurs in the data, base rate will be 0")
            theta[i, i] = -1e10
        else:
            theta[i, i] = np.log(sum_of_occurance / (state_container.data_size - sum_of_occurance + 1e-10))

    return np.around(theta, decimals=2)


"""
This part of the package contains functions related to (differentiated) uniformization as well as functions
to compute the log-likelihood score and its gradient for datasets that contain samples with known ages.

(see Rupp et al.(2021): 'Differentiated uniformization: A new method for inferring Markov chains on combinatorial state spaces including stochastic epidemic models')
"""
# author(s): Kevin Rupp, Stefan Vocht

cimport cython

import warnings

from scipy.linalg.cython_blas cimport dcopy, dscal, daxpy, ddot, dnrm2, dasum
from libc.stdlib cimport malloc, free
from libc.math cimport exp, log

from mhn.ssr.state_containers cimport State, StateContainer, StateAgeContainer
from mhn.ssr.state_space_restriction cimport get_mutation_num, restricted_q_vec, restricted_q_diag

import numpy as np
cimport numpy as np
from numpy.math cimport INFINITY

np.import_array()


@cython.wraparound(False)
@cython.boundscheck(False)
cdef void restricted_derivative_ik(double[:, :] theta_mat, int i, double[:] x_vec, State *state, int mutation_num, int k,
                                double *pout) nogil:
    """
    This function multiplies the kronecker product for the partial derivative of Q wrt to theta_ik with a vector

    :param theta_mat: matrix containing the theta entries
    :param i: vector is multiplied with the ith kronecker product (ith summand in eq. 9 of the original paper)
    :param x_vec: vector that is multiplied with the kronecker product
    :param state: current state used to compute the gradient
    :param mutation_num: number of mutations present in state
    :param k: column index of derivative
    :param pout: vector which will contain the result of this multiplication
    """

    # initialize some constants used in this function
    cdef double[:] theta_i = theta_mat[i, :]
    cdef int n = theta_i.shape[0]
    cdef int nx = 1 << mutation_num
    cdef int nxhalf = nx / 2
    cdef double mOne = -1
    cdef double zero = 0

    cdef int incx = 1
    cdef int incx2 = 2
    cdef int j

    # if the kth gene is not mutated and i != k (theta_ik no base rate), then the derivative is always zero
    if i != k and not (state[0].parts[k >> 5] >> (k & 31)) & 1:
        dscal(&nx, &zero, pout, &incx)
        return

    cdef double *ptmp = <double *> malloc(nx * sizeof(double))
    cdef double *px1
    cdef double *px2
    cdef double *shuffled_vec
    cdef double *old_vec
    cdef double *swap_vec
    cdef double theta

    # for the shuffle algorithm we have to initialize the pointers correctly
    if mutation_num & 1 == 1:
        swap_vec = ptmp
        shuffled_vec = pout
    else:
        swap_vec = pout
        shuffled_vec = ptmp

    old_vec = &x_vec[0]

    cdef int state_copy = state[0].parts[0]

    # use the shuffle algorithm to compute the product of the kronecker product with a vector
    for j in range(n):
        if state_copy & 1:
            dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
            dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)

            theta = exp(theta_i[j])
            px1 = shuffled_vec
            px2 = shuffled_vec + nxhalf
            # this function is only needed for the dua where we never have to transpose Q' or delete the diagonal
            if j == i:
                dcopy(&nxhalf, px1, &incx, px2, &incx)
                dscal(&nxhalf, &theta, px2, &incx)
                dcopy(&nxhalf, px2, &incx, px1, &incx)
                dscal(&nxhalf, &mOne, px1, &incx)

            else:
                dscal(&nxhalf, &theta, px2, &incx)
                if j == k:
                    dscal(&nxhalf, &zero, px1, &incx)

            old_vec = shuffled_vec
            shuffled_vec = swap_vec
            swap_vec = old_vec

        elif i == j:
            theta = -exp(theta_i[j])

            # if old_vec is still pointing to x_vec, we have to change it to not alter x_vec
            if old_vec == &x_vec[0]:
                dcopy(&nx, old_vec, &incx, swap_vec, &incx)
                old_vec = swap_vec

            dscal(&nx, &theta, old_vec, &incx)

		# if the mutation state of the next gene is stored on the current state_copy, make a bit shift to the right
		# else state_copy becomes the next integer stored in the given state (x >> 5  <=> x // 32, x & 31 <=> x % 32)
        if (j + 1) & 31:
            state_copy >>= 1
        else:
            state_copy = state[0].parts[(j+1) >> 5]

    free(ptmp)

@cython.wraparound(False)
@cython.boundscheck(False)
cdef void restricted_derivative_ik_diag(double[:, :] theta_mat, int i, State *state, int mutation_num, int k,
                                double *pout) except *:
    """
    This function calculates the diagonal of dQ/d theta_ik

    :param theta_mat: matrix containing the theta entries
    :param i: vector is multiplied with the ith kronecker product (ith summand in eq. 9 of the original paper)
    :param state: current state used to compute the gradient
    :param mutation_num: number of mutations present in state
    :param k: column index of derivative
    :param pout: vector which will contain the result of this multiplication
    """

    # initialize some constants used in this function
    cdef double[:] theta_i = theta_mat[i, :]
    cdef int n = theta_i.shape[0]
    cdef int nx = 1 << mutation_num
    cdef int nxhalf = nx / 2
    cdef double mOne = -1
    cdef double zero = 0

    cdef int incx = 1
    cdef int incx2 = 2
    cdef int j

    # if the kth gene is not mutated and i != k (theta_ik no base rate), then the derivative is always zero
    if i != k and not (state[0].parts[k >> 5] >> (k & 31)) & 1:
        dscal(&nx, &zero, pout, &incx)
        return

    cdef double *ptmp = <double *> malloc(nx * sizeof(double))
    cdef double *px1
    cdef double *px2
    cdef double *shuffled_vec
    cdef double *old_vec
    cdef double *swap_vec
    cdef double theta
    cdef double[:] x_vec = np.ones(nx, dtype=np.double)

    # for the shuffle algorithm we have to initialize the pointers correctly
    if mutation_num & 1 == 1:
        swap_vec = ptmp
        shuffled_vec = pout
    else:
        swap_vec = pout
        shuffled_vec = ptmp

    old_vec = &x_vec[0]

    cdef int state_copy = state[0].parts[0]

    # use the shuffle algorithm to compute the product of the kronecker product with a vector
    for j in range(n):
        if state_copy & 1:
            dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
            dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)

            theta = exp(theta_i[j])
            px1 = shuffled_vec
            px2 = shuffled_vec + nxhalf
            # this function is only needed for the dua where we never have to transpose dQ or remove its diagonal
            if j == i:
                dcopy(&nxhalf, px1, &incx, px2, &incx)
                dscal(&nxhalf, &theta, px2, &incx)
                dcopy(&nxhalf, px2, &incx, px1, &incx)
                dscal(&nxhalf, &mOne, px1, &incx)
                dscal(&nxhalf, &zero, px2, &incx)

            else:
                dscal(&nxhalf, &theta, px2, &incx)
                if j == k:
                    dscal(&nxhalf, &zero, px1, &incx)

            old_vec = shuffled_vec
            shuffled_vec = swap_vec
            swap_vec = old_vec

        elif i == j:
            theta = -exp(theta_i[j])

            # if old_vec is still pointing to x_vec, we have to change it to not alter x_vec
            if old_vec == &x_vec[0]:
                dcopy(&nx, old_vec, &incx, swap_vec, &incx)
                old_vec = swap_vec

            dscal(&nx, &theta, old_vec, &incx)

		# if the mutation state of the next gene is stored on the current state_copy, make a bit shift to the right
		# else state_copy becomes the next integer stored in the given state (x >> 5  <=> x // 32, x & 31 <=> x % 32)
        if (j + 1) & 31:
            state_copy >>= 1
        else:
            state_copy = state[0].parts[(j+1) >> 5]

    free(ptmp)


@cython.wraparound(False)
@cython.boundscheck(False)
cdef double[:] cython_restricted_expm(double[:, :] theta, double[:] b, State *state, double t, double eps) except *:
    """
    this functions multiplies expm(tQ) with a vector b

    :param theta: matrix containing the theta entries
    :param b: array that is multiplied with expm(tQ)
    :param state: state representing current tumor sample
    :param t: age of state
    :param eps: accuracy
    """
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int i_one = 1

    # Compute the diagonal of Q
    cdef double *dg = <double *> malloc(nx * sizeof(double))
    restricted_q_diag(theta, state, dg)
    # Calculate the L2-norm of the diagonal to use as scaling constant gamma
    cdef double gam =  dnrm2(&nx, dg, &i_one)
    cdef double gam_inv = 1/gam

    cdef np.ndarray[np.double_t] pt = np.zeros(nx, dtype=np.double)
    cdef int n = 0
    cdef double w = 1.0
    cdef double egtw = exp(-1.0*gam*t)
    cdef double *q_vec_result = <double *> malloc(nx * sizeof(double))
    cdef double[:] q = b.copy()
    cdef double mass_defect = 0.0
    cdef double b_sum = np.sum(b)

    while eps < b_sum * (1 - mass_defect):
        mass_defect += egtw
        daxpy(&nx, &egtw, &q[0], &i_one, &pt[0], &i_one)
        n += 1
        # Calculate q = [1/gamma*Q+I]b
        restricted_q_vec(theta, q, state, q_vec_result, diag=True, transp=False) # q=1/gamma*Qb
        daxpy(&nx, &gam_inv, q_vec_result, &i_one, &q[0], &i_one) # calculate q=q+Ib
        egtw *= gam*t/n

    free(dg)
    free(q_vec_result)
    return pt


@cython.wraparound(False)
@cython.boundscheck(False)
def restricted_expm(double[:, :] theta, double[:] b, StateAgeContainer state_with_age, double eps):
    """
    this functions multiplies expm(tQ) with a vector b, this is a Python wrapper for the internal function cython_restricted_expm

    :param theta: matrix containing the theta entries
    :param b: array that is multiplied with expm(tQ)
    :param state_with_age: StateAgeContainer containing exactly one state with its corresponding age
    :param eps: accuracy
    """
    if state_with_age.data_size != 1:
        raise ValueError("state_with_age is expected to contain exactly one state with its corresponding age")
    cdef State *state = &state_with_age.states[0]
    cdef double t = state_with_age.state_ages[0]
    return np.asarray(cython_restricted_expm(theta, b, state, t, eps))


@cython.wraparound(False)
@cython.boundscheck(False)
cdef calc_gamma(double[:, :] theta, State *state, int i, int k):
    """
    this function calculates the derivative of the scaling factor gamma wrt. theta_ik
    :param theta: matrix containing the theta entries
    :param state: state representing current tumor sample
    :param i: row index of theta entry to take derivative wrt.
    :param k: column index of theta entry to take derivative wrt.
    """
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int one = 1
    cdef double *deriv_q_diag = <double *> malloc(nx * sizeof(double))
    restricted_derivative_ik_diag(theta, i, state, mutation_num, k, deriv_q_diag)
    cdef double *q_diag = <double*> malloc(nx*sizeof(double))
    restricted_q_diag(theta, state, q_diag)
    cdef double num = ddot(&nx, deriv_q_diag, &one, q_diag, &one)
    cdef double denom = dnrm2(&nx, q_diag, &one)
    free(deriv_q_diag)
    free(q_diag)
    return denom, num / denom


def calc_gamma_wrapper(double[:, :] theta, StateContainer state, int i, int k):
    """
    this function calculates the derivative of the scaling factor gamma wrt. theta_ik. It is a Python wrapper for the
    internal Cython function to compute gamma and its derivative.

    :param theta: matrix containing the theta entries
    :param state: a StateContainer that contains one state for which gamma will be computed
    :param i: row index of theta entry to take derivative wrt.
    :param k: column index of theta entry to take derivative wrt.
    :return: gamma and its derivative wrt. theta_ik
    """
    return calc_gamma(theta, state.states, i, k)


@cython.wraparound(False)
@cython.boundscheck(False)
cdef void dua(double[:, :] theta, double[:] b, State *state, double t, int i, int k, double eps, double[:] pt, double[:] dp) except *:
    """
    Computes the frechet derivative of expm(tQ)b using the DUA Algorithm
    :param theta: matrix containing the theta entries
    :param b: distribution to be multiplied from the right
    :param state: state representing the current tumor sample
    :param t: Age of tumor, amount of time to be projected in the future
    :param i: row index of theta to take the derivative wrt.
    :param k: column index of theta to take the derivative wrt.
    :param eps: accuracy
    :param pt: container to store the resulting pt
    :param dp: container to store the derivative wrt. theta_ik
    """
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int j

    cdef double w = 1.0
    cdef int n = 0 # Iteration number
    cdef int one = 1
    cdef double done = 1.0
    cdef double zero = 0
    cdef double gfac = 1.0
    dscal(&nx, &zero, &pt[0], &one)
    dscal(&nx, &zero, &dp[0], &one)
    cdef double[:] q = b.copy()
    cdef double[:] dq = np.zeros(nx, dtype=np.double)
    cdef double * temp = <double *> malloc(nx * sizeof(double))
    cdef double * temp2 = <double *> malloc(nx * sizeof(double))

    cdef double gamma, dgamma
    gamma, dgamma = calc_gamma(theta, state, i, k)
    cdef double dgam_inv = -1.0/gamma**2*dgamma
    cdef double gam_inv = 1/gamma
    cdef double ewg = exp(-1.0*gamma*t)
    cdef double mass_defect = 0.0
    while eps < (1 - mass_defect):
        mass_defect += ewg
        # pt = pt + exp(-gam*t)q
        daxpy(&nx, &ewg, &q[0], &one, &pt[0], &one)
        # dpt = dpt + exp(-gamma*t)w dq
        daxpy(&nx, &ewg, &dq[0], &one, &dp[0], &one)
        # dpt = dpt + exp(-gamma*t)w dgamma(n/gamma-t)q
        gfac = ewg*dgamma*(n/gamma-t)
        daxpy(&nx, &gfac, &q[0], &one, &dp[0], &one)

        n += 1
        # dq = -1/gamma^2*dg*Q q + 1/gamma*dQ q
        restricted_q_vec(theta, q, state, temp, True, False)
        dscal(&nx, &dgam_inv, temp, &one)
        restricted_derivative_ik(theta, i, q, state, mutation_num, k, temp2)
        daxpy(&nx, &gam_inv, temp2, &one, temp, &one) # temp2 isn't needed anymore and its allocated memory can be reused

        # dq = dq + [1/gamma*Q+I]dq
        restricted_q_vec(theta, dq, state, temp2, True, False)
        daxpy(&nx, &gam_inv, temp2, &one, &dq[0], &one)
        for j in range(nx):
            dq[j] += temp[j] # +temp2[j]

        # q = [1/gamma*Q + I]q
        restricted_q_vec(theta, q, state, temp, True, False)
        daxpy(&nx, &gam_inv, temp, &one, &q[0], &one)

        ewg *= gamma*t/n

    free(temp)
    free(temp2)


cdef bint compute_diff_state(State *former_state, State *current_state, State *diff_state):
    """
    Find all mutations present in the current state that are not in the former state
    
    :param former_state: state of the previous sample
    :param current_state: state of the current sample
    :param diff_state: state that contains only the mutations present in current sample that are not in former sample
    
    :returns: False, if former_state contains a mutations current_state does not contain -> error
    """

    cdef int i
    cdef int former_state_copy, current_state_copy

    for i in range(STATE_SIZE):
        former_state_copy = former_state[0].parts[i]
        current_state_copy = current_state[0].parts[i]
        # check if there is a mutation in the former state that is not in the current state
        if (current_state_copy - former_state_copy) != (current_state_copy ^ former_state_copy):
            return False
        diff_state[0].parts[i] = (current_state_copy ^ former_state_copy)

    return True


cdef compute_modified_theta(double[:, :] theta, double[:, :] modified_theta, State *former_state):
    """
    Modifies theta according to the mutations present in the former sample.
    It sets the base rates of genes that are mutated in the former sample to 0 (this means -inf in log space) and 
    adds the multiplicative effects of those mutations to the base rates of genes that are not mutated in the previous
    sample.
    
    :param theta: actual theta matrix
    :param modified_theta: array that will contain the modified theta at the end
    :param former_state: State object representing the previous sample which is used to compute the modified theta
    """
    cdef int n = theta.shape[0]
    cdef int n_square = n*n
    cdef int one = 1
    cdef int i, j
    cdef int state_copy_i = former_state[0].parts[0]
    cdef int state_copy_j

    # initialize the modified theta with the actual theta matrix
    dcopy(&n_square, &theta[0, 0], &one, &modified_theta[0, 0], &one)

    # look for genes that are mutated in the previous sample and modify all base rates accordingly
    for i in range(n):
        if state_copy_i & 1:
            state_copy_j = former_state[0].parts[0]
            # we want exp(theta_ii) to be 0 if i was mutated in former state
            modified_theta[i, i] = -INFINITY
            # all base rates are amplified by i as i is always mutated
            for j in range(n):
                if not state_copy_j & 1:
                    modified_theta[j, j] += theta[j, i]

                if (j + 1) & 31:
                    state_copy_j >>= 1
                else:
                    state_copy_j = former_state[0].parts[(j + 1) >> 5]

        if (i + 1) & 31:
            state_copy_i >>= 1
        else:
            state_copy_i = former_state[0].parts[(i + 1) >> 5]


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef cython_gradient_and_score(double[:, :] theta, StateAgeContainer mutation_data, double eps):
    """
    This function computes the log-likelihood score and its gradient for a given theta and data, where we know the ages
    of the individual samples (see Rupp et al. (2021) eq. (13)-(15))

    :param theta: theta matrix representing the MHN
    :param mutation_data: data from which we learn the MHN
    :param eps: accuracy
    :returns: the gradient and the score as a tuple
    """
    cdef int n = theta.shape[0]
    cdef int current_nx
    cdef int data_size = mutation_data.data_size
    cdef int k, i, j
    cdef double t
    cdef double score = 0
    cdef double diagonal_partial_grad
    cdef np.ndarray[np.double_t, ndim=2] gradient = np.zeros((n, n), dtype=np.double)
    cdef np.ndarray[np.double_t, ndim=2] modified_theta = np.empty((n, n))
    cdef np.ndarray[np.double_t] pt
    cdef np.ndarray[np.double_t] dp
    cdef np.ndarray[np.double_t] b
    cdef State diff_state
    cdef int state_copy, state_copy_former

    for k in range(1, data_size):
        # sample k must always be older than sample k-1
        t = mutation_data.state_ages[k] - mutation_data.state_ages[k-1]
        assert t >= 0

        # get the mutations that are new in the current sample and make sure that no mutation has disappeared
        if not compute_diff_state(&mutation_data.states[k-1], &mutation_data.states[k], &diff_state):
            warnings.warn(f"The sample at position {k} contains less mutations than the previous sample")
            continue

        # as we restrict the state-space to contain only states that are "between" the previous and the current state
        # this means that we get the score by multiplying exp(Qt) with (1, 0, ..., 0) and taking the last value
        # (index = current_nx-1) of the resulting vector
        current_nx = 1 << get_mutation_num(&diff_state)
        compute_modified_theta(theta, modified_theta, &mutation_data.states[k-1])
        # For the unlikely case that we have two samples with all genes mutated, the dua algorithm would never be
        # executed. In that case we need pt to be 1 for the score computation at the end, so we initialize pt with ones.
        # If the dua function is called even once, then all pt entries are set to zero there anyway, so this init is fine.
        pt = np.ones(current_nx)
        dp = np.empty(current_nx)
        b = np.zeros(current_nx)
        b[0] = 1
        state_copy_former = mutation_data.states[k-1].parts[0]
        for i in range(n):
            # if the gene was mutated in the previous sample, row i of the gradient matrix is zero everywhere
            if not state_copy_former & 1:
                state_copy = diff_state.parts[0]
                # the gradient of theta_ij is non-zero if gene j is mutated in the current sample
                for j in range(n):
                    if state_copy & 1 or i == j:
                        dua(modified_theta, b, &diff_state, t, i, j, eps, pt, dp)
                        gradient[i, j] += dp[current_nx-1] / pt[current_nx-1]
                        if i == j:
                            diagonal_partial_grad = dp[current_nx-1] / pt[current_nx-1]

                    if (j + 1) & 31:
                        state_copy >>= 1
                    else:
                        state_copy = diff_state.parts[(j + 1) >> 5]

                # if gene j was mutated in the previous sample then the gradient of theta_ij is the same as theta_ii
                state_copy = mutation_data.states[k-1].parts[0]
                for j in range(n):
                    if state_copy & 1:
                        gradient[i, j] += diagonal_partial_grad

                    if (j + 1) & 31:
                        state_copy >>= 1
                    else:
                        state_copy = mutation_data.states[k-1].parts[(j + 1) >> 5]

            if (i + 1) & 31:
                state_copy_former >>= 1
            else:
                state_copy_former = mutation_data.states[k-1].parts[(i + 1) >> 5]

        score += log(pt[current_nx-1])

    return gradient, score


# author(s): Stefan Vocht
#
# this .pxd file is the cython header file for state_containers.pyx
#


# STATE_SIZE is defined in setup.py, 
# the maximum number n of genes the MHN can handle is 32 * STATE_SIZE
cdef extern from *:
    """
    typedef struct {
        uint32_t parts[STATE_SIZE];
    } State;
    """
    ctypedef struct State:
        unsigned int parts[STATE_SIZE]


cdef class StateContainer:
    cdef State *states
    cdef int data_size
    cdef int gene_num
    cdef int max_mutation_num


cdef class StateAgeContainer(StateContainer):
    cdef double *state_ages







from . import optimizers
from . import model
from . import ssr
from . import original

from .ssr.state_space_restriction import cuda_available
from .ssr.state_space_restriction import CUDA_AVAILABLE, CUDA_NOT_AVAILABLE, CUDA_NOT_FUNCTIONAL



"""
This submodule implements UtilityFunctions.R from the original R implementation in Python.

It contains functions useful for preprocessing training data.
"""
# author(s): Stefan Vocht

import numpy as np


def state_to_int(x: np.ndarray) -> int:
    """
    This function interprets a binary array x as a binary number and returns the corresponding value as an integer.

    :param x: binary array, in the context of the script this is a row of the mutation matrix
    :return: integer number representing the binary array
    """

    # reverse list and convert elements to string
    x = map(str, x[::-1])
    return int(''.join(x), 2)


def data_to_pD(data: np.ndarray) -> np.ndarray:
    """
    This function calculates the probability distribution for the different events for a given binary mutation matrix.

    :param data: has to be a numpy array/matrix (mutation matrix)
    :return: probability distribution of the different events
    """

    n = data.shape[1]
    N = 2**n

    # convert data into a list of integers, where each number represents a different event
    data = list(map(state_to_int, data))

    # calculate the probability distribution
    pD = np.bincount(data, minlength=N)
    pD = pD / np.sum(pD)

    return pD


def finite_sample(pTh: np.ndarray, k: int) -> np.ndarray:
    """
    Generates a random sample given a probability distribution and returns the probability distribution for this new
    sample.

    :param pTh: probability distribution of events (the distribution of a true Theta)
    :param k: number of samples that should be generated
    :return: probability distribution of events from the generated samples
    """
    n = pTh.size

    return np.bincount(np.random.choice(n, k, replace=True, p=pTh), minlength=n) / k


def KL_div(p: np.ndarray, q: np.ndarray) -> float:
    """
    Computes the Kullback–Leibler divergence of two probability distributions.

    :param p: probability distribution p
    :param q: probability distribution q
    :return: KL-divergence of p and q
    """
    return p.dot(np.log(p)) - p.dot(np.log(q))



"""
This submodule contains functions to work with the transition rate matrix Q in a restricted state-space as well as
functions to compute the marginal log-likelihood score and its gradient making use of state-space restriction.
"""
# author(s): Stefan Vocht

cimport cython

from scipy.linalg.cython_blas cimport dcopy, dscal, daxpy, ddot, dnrm2
from libc.stdlib cimport malloc, free
from libc.math cimport exp, log

from mhn.ssr.state_containers cimport State, StateContainer
from mhn.original.PerformanceCriticalCode cimport _compute_inverse, _compute_inverse_t

import numpy as np
cimport numpy as np

np.import_array()

cdef extern from *:
    """
    /*
    Counts number of 1s in binary representation of number x, where x is a 32-bit integer
    Source: https://stackoverflow.com/questions/109023/how-to-count-the-number-of-set-bits-in-a-32-bit-integer
    */
    int count_ones32(uint32_t i){
        i = i - ((i >> 1) & 0x55555555);                                    // add pairs of bits
        i = (i & 0x33333333) + ((i >> 2) & 0x33333333);                     // quads
        i = (i + (i >> 4)) & 0x0F0F0F0F;                                    // groups of 8
        return (i * 0x01010101) >> 24;                                      // horizontal sum of bytes
    }

    /*
    Counts number of 1s in binary representation of number x, where x is a 64-bit integer
    Source: https://en.wikipedia.org/wiki/Hamming_weight 
    */
    int count_ones(long long x) {
        x -= (x >> 1) & 0x5555555555555555LL;             					//put count of each 2 bits into those 2 bits
        x = (x & 0x3333333333333333LL) + ((x >> 2) & 0x3333333333333333LL); //put count of each 4 bits into those 4 bits 
        x = (x + (x >> 4)) & 0x0f0f0f0f0f0f0f0fLL;        					//put count of each 8 bits into those 8 bits 
        return (x * 0x0101010101010101LL) >> 56;  							//returns left 8 bits of x + (x<<8) + (x<<16) + (x<<24) + ... 
    }
    """
    inline int count_ones32(unsigned int u) nogil
    inline int count_ones(long long x) nogil


def count_ones64(long long x):
    """
    Counts the number of bits set to 1 in a 64-bit integer.
    Wrapper so that count_ones can be called from a Python script
    """
    return count_ones(x)


cdef int get_mutation_num(State *state):
    """
    Get the number of mutations in a given state
    """
    cdef int mutation_num = 0
    cdef int i

    for i in range(STATE_SIZE):
        mutation_num += count_ones32(state[0].parts[i])

    return mutation_num


# load the function cuda_gradient_and_score if the CUDA compiler is available
IF NVCC_AVAILABLE:
    cdef extern from *:
        """
        #ifdef _WIN32
        #define DLL_PREFIX __declspec(dllexport)
        #else
        #define DLL_PREFIX 
        #endif

        int DLL_PREFIX cuda_gradient_and_score_implementation(double *ptheta, int n, State *mutation_data, int data_size, double *grad_out, double *score_out);
        void DLL_PREFIX get_error_name_and_description(int error, const char **error_name, const char **error_description);
        int DLL_PREFIX cuda_functional();
        """

        int cuda_gradient_and_score_implementation(double *ptheta, int n, State *mutation_data, int data_size, double *grad_out, double *score_out)
        void get_error_name_and_description(int error, const char **error_name, const char **error_description)
        int cuda_functional()


@cython.wraparound(False)
@cython.boundscheck(False)
cdef void restricted_kronvec(double[:, :] theta_mat, int i, double[:] x_vec, State *state, int mutation_num, double *pout, bint diag = False, bint transp = False) nogil:
    """
    This function multiplies the kronecker product described in the original MHN paper in eq.9 with a vector

    :param theta_mat: matrix containing the theta entries
    :param i: vector is multiplied with the ith kronecker product (ith summand in eq. 9 of the original paper) 
    :param x_vec: vector that is multiplied with the kronecker product
    :param state: current state used to compute the gradient
    :param mutation_num: number of mutations present in state
    :param pout: vector which will contain the result of this multiplication
    :param diag: if False, the diagonal of the kronecker product is set to zero
    :param transp: if True, the kronecker product is transposed
    """

    # initialize some constants used in this function
    cdef double[:] theta_i = theta_mat[i, :]
    cdef int n = theta_i.shape[0]
    cdef int nx = 1 << mutation_num
    cdef int nxhalf = nx / 2
    cdef double mOne = -1
    cdef double zero = 0

    cdef int incx = 1
    cdef int incx2 = 2
    cdef int j

    # if we have no diagonal and the ith gene is not mutated, the result is always a zero vector
    if not diag and not (state[0].parts[i >> 5] >> (i & 31)) & 1:
        dscal(&nx, &zero, pout, &incx)
        return

    cdef double *ptmp = <double *> malloc(nx * sizeof(double))
    cdef double *px1
    cdef double *px2
    cdef double *shuffled_vec
    cdef double *old_vec
    cdef double *swap_vec
    cdef double theta

    # for the shuffle algorithm we have to initialize the pointers correctly
    if mutation_num & 1 == 1:
        swap_vec = ptmp
        shuffled_vec = pout
    else:
        swap_vec = pout
        shuffled_vec = ptmp

    old_vec = &x_vec[0]

    cdef int state_copy = state[0].parts[0]

    # use the shuffle algorithm to compute the product of the kronecker product with a vector
    for j in range(n):
        if state_copy & 1:
            dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
            dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)

            theta = exp(theta_i[j])
            px1 = shuffled_vec
            px2 = shuffled_vec + nxhalf

            if j == i:
                if not transp:
                    dcopy(&nxhalf, px1, &incx, px2, &incx)
                    dscal(&nxhalf, &theta, px2, &incx)
                    if diag:
                        dcopy(&nxhalf, px2, &incx, px1, &incx)
                        dscal(&nxhalf, &mOne, px1, &incx)
                    else:
                        dscal(&nxhalf, &zero, px1, &incx)
                else:
                    if diag:
                        theta *= -1
                        daxpy(&nxhalf, &mOne, px2, &incx, px1, &incx)
                        dscal(&nxhalf, &theta, px1, &incx)
                        dscal(&nxhalf, &zero, px2, &incx)
                    else:
                        dcopy(&nxhalf,px2,&incx,px1,&incx)
                        dscal(&nxhalf,&theta,px1,&incx)
                        dscal(&nxhalf,&zero,px2,&incx)

            else:
                dscal(&nxhalf, &theta, px2, &incx)
 
            old_vec = shuffled_vec
            shuffled_vec = swap_vec
            swap_vec = old_vec

        elif i == j:
            theta = -exp(theta_i[j])

            # if old_vec is still pointing to x_vec, we have to change it to not alter x_vec
            if old_vec == &x_vec[0]:
                dcopy(&nx, old_vec, &incx, swap_vec, &incx)
                old_vec = swap_vec

            dscal(&nx, &theta, old_vec, &incx)

		# if the mutation state of the next gene is stored on the current state_copy, make a bit shift to the right
		# else state_copy becomes the next integer stored in the given state (x >> 5  <=> x // 32, x & 31 <=> x % 32)
        if (j + 1) & 31:
            state_copy >>= 1
        else:
            state_copy = state[0].parts[(j+1) >> 5]

    free(ptmp)


cdef void restricted_q_vec(double[:, :] theta, double[:] x, State *state, double *yout, bint diag= False, bint transp = False):
    """
    multiplies the matrix Q(ptheta) with the vector x, result is saved in yout

    :param theta: matrix containing the theta entries
    :param x: vector that should be multiplied with Q(ptheta)
    :param state: state representing current tumor sample
    :param yout: array in which the result is stored
    :param diag: if False, the diag of Q is set to zero during multiplication
    :param transp: if True, multiplication is done with the transposed Q
    """

    cdef int n = theta.shape[0]
    cdef int nx = x.shape[0]
    cdef int i
    cdef double one_d = 1
    cdef int one_i = 1
    cdef double zero = 0

    # get the number of mutations present in the given state
    cdef int mutation_num = get_mutation_num(state)

    cdef double *result_vec = <double *> malloc(sizeof(double) * nx)

    # initialize yout with zero
    dscal(&nx, &zero, yout, &one_i)

    for i in range(n):
        restricted_kronvec(theta, i, x, state, mutation_num, result_vec, diag, transp)
        # add result of restricted_kronvec to yout
        daxpy(&nx, &one_d, result_vec, &one_i, yout, &one_i)

    free(result_vec)


@cython.wraparound(False)
@cython.boundscheck(False)
cdef void restricted_q_diag(double[:, :] theta, State *state, double *dg):
    """
    Compute the diagonal of the transition rate matrix Q

    :param theta: matrix containing the theta entries
    :param state: state representing current tumor sample
    :param dg: array in which the diagonal is stored at the end
    """
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int n = theta.shape[0]
    cdef int i, j
    cdef int state_copy

    cdef double *s = <double *> malloc(nx * sizeof(double))
    cdef int current_length
    cdef double exp_theta
    cdef double d_one = 1
    cdef double zero = 0
    cdef int i_one = 1

    # initialize the diagonal with zero
    dscal(&nx, &zero, dg, &i_one)

    # compute the ith subdiagonal of Q and add it to dg
    for i in range(n):
        state_copy = state[0].parts[0]
        current_length = 1
        s[0] = 1
        # compute the ith subdiagonal of Q 
        for j in range(n):
            if state_copy & 1:
                exp_theta = exp(theta[i, j])
                if i == j:
                    exp_theta *= -1
                    dscal(&current_length, &exp_theta, s, &i_one)
                    dscal(&current_length, &zero, s + current_length, &i_one)
                else:
                    dcopy(&current_length, s, &i_one, s + current_length, &i_one)
                    dscal(&current_length, &exp_theta, s + current_length, &i_one)

                current_length *= 2

            elif i == j:
                exp_theta = - exp(theta[i, j])
                dscal(&current_length, &exp_theta, s, &i_one)

            # if the mutation state of the next gene is stored on the current state_copy, make a bit shift to the right
            # else state_copy becomes the next integer stored in the given state (x >> 5  <=> x // 32, x & 31 <=> x % 32)
            if (j + 1) & 31:
                state_copy >>= 1
            else:
                state_copy = state[0].parts[(j+1) >> 5]

        # add the subdiagonal to dg
        daxpy(&nx, &d_one, s, &i_one, dg, &i_one)

    free(s)


@cython.wraparound(False)
@cython.boundscheck(False)
cdef np.ndarray[np.double_t] restricted_jacobi(double[:, :] theta, double[:] b, State *state, bint transp = False):
    """
    this functions multiplies [I-Q]^(-1) with b

    :param theta: matrix containing the theta entries
    :param b: array that is multiplied with [I-Q]^(-1)
    :param state: state representing current tumor sample
    :param transp: if True, b is multiplied with the transposed [I-Q]^(-1)
    """
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int z, j
    cdef int i_one = 1
    cdef int zero = 0
    cdef double d_one = 1
    cdef double mOne = -1

    # make x a vector of size nx in which every entry is set to 1/nx,
    # will be the initial guess to the solution for the jacobi method
    cdef np.ndarray[np.double_t] x = np.full(nx, 1 / (1.0 * nx), dtype=np.double)
    cdef double *q_vec_result = <double *> malloc(nx * sizeof(double))

    # compute the diagonal of [I-Q], store it in dg
    cdef double *dg = <double *> malloc(nx * sizeof(double))
    restricted_q_diag(theta, state, dg)             # compute the diagonal of Q
    daxpy(&nx, &d_one, &mOne, &zero, dg, &i_one)    # subtract 1 from each entry to get the diagonal of [Q-I]
    dscal(&nx, &mOne, dg, &i_one)                   # scale with -1 to get the diagonal of [I-Q]

    for z in range(mutation_num+1):
        restricted_q_vec(theta, x, state, q_vec_result, diag=False, transp=transp)
        # add b to the result of q_vec
        daxpy(&nx, &d_one, &b[0], &i_one, q_vec_result, &i_one)
        # divide every entry by its corresponding diagonal entry
        for j in range(nx):
            x[j] = q_vec_result[j] / dg[j]

    free(dg)
    free(q_vec_result)
    return x


cdef void _compute_restricted_inverse(double[:, :] theta, double *dg, State *state, double[:] b, double[:] xout, bint transp = False):
    """
    this functions multiplies [I-Q]^(-1) with b and is much faster than restricted jacobi
    
    :param theta: matrix containing the theta entries
    :param dg: vector containing the diagonal of [I-Q]
    :param state: state representing current tumor sample
    :param b: array that is multiplied with [I-Q]^(-1)
    :param xout: vector which will contain the result at the end
    :param transp: if True, b is multiplied with the transposed [I-Q]^(-1)
    """
    cdef int n = theta.shape[0]
    cdef int mutation_num = get_mutation_num(state)
    cdef int i, j
    cdef int state_copy

    cdef double *mutated_thetas = <double *> malloc(mutation_num * mutation_num * sizeof(double))
    cdef int *mutation_pos = <int *> malloc(mutation_num * sizeof(int))

    # we can use the compute_inverse functions from the full state-space code
    # we only have to modify theta, such that the theta that we pass to the compute_inverse functions
    # only contains the thetas that correspond to mutated genes in state

    # first get the indices of the mutated genes
    j = 0
    state_copy = state[0].parts[0]
    for i in range(n):
        if state_copy & 1:
            mutation_pos[j] = i
            j += 1
        if (i + 1) & 31:
            state_copy >>= 1
        else:
            state_copy = state[0].parts[(i+1) >> 5]

    # get only the thetas that correspond to mutated genes
    for i in range(mutation_num):
        for j in range(mutation_num):
            mutated_thetas[i*mutation_num + j] = theta[mutation_pos[i], mutation_pos[j]]

    # now simply call the compute_inverse functions with the "mutated" thetas and pass mutation_num instead of n
    if transp:
        _compute_inverse_t(mutated_thetas, mutation_num, dg, &b[0], &xout[0])
    else:
        _compute_inverse(mutated_thetas, mutation_num, dg, &b[0], &xout[0])

    free(mutation_pos)
    free(mutated_thetas)


@cython.wraparound(False)
@cython.boundscheck(False)
def compute_restricted_inverse(double[:, :] theta, int[:] state, double[:] b, bint transp = False) -> np.ndarray:
    """
    This functions multiplies [I-Q]^(-1) with b, where Q is the restricted transition rate matrix constructed from the given state.
    This is a Python wrapper for the internal Cython function.

    :param theta: matrix containing the theta entries
    :param state: state representing current tumor sample, has to be a binary 1d numpy array with dtype=np.int32, where each entry corresponds to an event being present or not
    :param b: array that is multiplied with [I-Q]^(-1)
    :param transp: if True, b is multiplied with the transposed [I-Q]^(-1)

    :returns: the result as numpy array
    """
    cdef State c_state
    cdef int nx = b.shape[0]
    cdef int j
    cdef int incx = 1, incx0 = 0
    cdef double mOne = -1, one = 1

    cdef np.ndarray[np.double_t] xout = np.empty(nx, dtype=np.double)
    cdef double *dg = <double *> malloc(nx * sizeof(double))

    for j in range(STATE_SIZE):
        c_state.parts[j] = 0

    for j in range(state.shape[0]):
        if state[j]:
            c_state.parts[j >> 5] |= 1 << (j & 31)

    restricted_q_diag(theta, &c_state, dg)      # compute the diagonal of Q
    daxpy(&nx, &one, &mOne, &incx0, dg, &incx)  # subtract 1 from each entry to get the diagonal of [Q-I]
    dscal(&nx, &mOne, dg, &incx)                # scale with -1 to get the diagonal of [I-Q]

    _compute_restricted_inverse(theta, dg, &c_state, b, xout, transp)

    free(dg)
    return xout


@cython.wraparound(False)
@cython.boundscheck(False)
cdef double restricted_gradient_and_score(double[:, :] theta, State *state, double[:, :] g):
    """
    Computes a part of the gradient and score corresponding to a given state

    :param theta: matrix containing the theta entries
    :param state: state representing current tumor sample
    :param g: the resulting gradient is stored in this matrix
    :return: part of the total score
    """
    cdef int n = theta.shape[0]
    cdef int mutation_num = get_mutation_num(state)
    cdef int nx = 1 << mutation_num
    cdef int nxhalf = nx / 2
    cdef int incx = 1
    cdef int incx2 = 2
    cdef int incx0 = 0
    cdef double one = 1.
    cdef double mOne = -1.
    cdef np.ndarray[np.double_t] p0 = np.zeros(nx, dtype=np.double)
    p0[0] = 1

    # compute dg, the diagonal of [I-Q]
    cdef double *dg = <double *> malloc(nx * sizeof(double))
    restricted_q_diag(theta, state, dg)         # compute the diagonal of Q
    daxpy(&nx, &one, &mOne, &incx0, dg, &incx)  # subtract 1 from each entry to get the diagonal of [Q-I]
    dscal(&nx, &mOne, dg, &incx)                # scale with -1 to get the diagonal of [I-Q]

    # compute parts of the probability distribution yielded by the current MHN
    cdef np.ndarray[np.double_t] pth = np.empty(nx, dtype=np.double)
    _compute_restricted_inverse(theta, dg, state, p0, pth)

    cdef np.ndarray[np.double_t] pD = np.zeros(nx)
    pD[nx-1] = 1 / pth[nx-1]

    cdef np.ndarray[np.double_t] q = np.empty(nx, dtype=np.double)
    _compute_restricted_inverse(theta, dg, state, pD, q, True)

    cdef int i, j

    for i in range(n):
        for j in range(n):
            g[i, j] = 0

    cdef int state_copy
    cdef double *r_vec = <double *> malloc(nx * sizeof(double))

    cdef double *shuffled_vec
    cdef double *old_vec
    cdef double *swap_vec

    cdef double *ptmp = <double *> malloc(nx * sizeof(double))

    # compute the gradient efficiently using the shuffle trick
    for i in range(n):
        restricted_kronvec(theta, i, pth, state, mutation_num, ptmp, diag=True)
        for j in range(nx):
            r_vec[j] = q[j] * ptmp[j] 
        old_vec = &r_vec[0]
        # reuse the allocated memory of ptmp for the shuffle since the entries of ptmp are no longer used anyway
        shuffled_vec = ptmp
        state_copy = state[0].parts[0]
        for j in range(n):
            if state_copy & 1:
                # shuffle entries of old_vec
                dcopy(&nxhalf, old_vec, &incx2, shuffled_vec, &incx)
                dcopy(&nxhalf, old_vec+1, &incx2, shuffled_vec+nxhalf, &incx)
                # add up the entries of the second half of the shuffled vector to get the partial derivative
                g[i, j] = ddot(&nxhalf, shuffled_vec+nxhalf, &incx, &one, &incx0)
                # in the case i == j also add all the entries of the first half of the shuffled vector
                if i == j:
                    g[i, j] += ddot(&nxhalf, shuffled_vec, &incx, &one, &incx0)

                # make shuffled_vec the old_vec and vice versa for the next iteration
                swap_vec = old_vec
                old_vec = shuffled_vec
                shuffled_vec = swap_vec

            elif i == j:
                # add up all entries of old_vec (respectively r_vec) to get the partial derivative for i == j, if gene i is not mutated
                g[i, j] = ddot(&nx, old_vec, &incx, &one, &incx0)

            # if the mutation state of the next gene is stored on the current state_copy, make a bit shift to the right
            # else state_copy becomes the next integer stored in the given state (x >> 5  <=> x // 32, x & 31 <=> x % 32)
            if (j + 1) & 31:
                state_copy >>= 1
            else:
                state_copy = state[0].parts[(j+1) >> 5]

    free(ptmp)
    free(r_vec)
    free(dg)

    return log(pth[nx - 1])


cpdef cython_gradient_and_score(double[:, :] theta, StateContainer mutation_data):
    """
    Computes the total gradient and score for a given MHN and given mutation data

    :param theta: matrix containing the theta entries of the current MHN
    :param mutation_data: StateContainer containing the mutation data the MHN should be trained on
    :return: tuple containing the gradient and the score
    """
    cdef int n = theta.shape[0]
    cdef int data_size = mutation_data.data_size
    cdef int i, j
    cdef np.ndarray[np.double_t, ndim=2] final_gradient = np.zeros((n, n), dtype=np.double)
    cdef double *local_grad_sum
    cdef np.ndarray[np.double_t, ndim=2] local_gradient_container = np.empty((n, n), dtype=np.double)
    cdef double zero = 0
    cdef int incx = 1
    cdef double one = 1
    cdef int n_square = n*n

    cdef double score = 0

    for i in range(data_size):
        # for each sample/patient in mutation_data,
        # compute the gradient and score for the sample and add them to the total gradient and total score
        score += restricted_gradient_and_score(theta, &mutation_data.states[i], local_gradient_container)
        final_gradient += local_gradient_container

    # return the normalized gradient and normalized score
    return (final_gradient / data_size), (score / data_size)


class CUDAError(Exception):
    """
    Error raised if something went wrong during execution of the CUDA code.
    """


cpdef cuda_gradient_and_score(double[:, :] theta, StateContainer mutation_data):
    """
    This function is a wrapper for the CUDA implementation of state-space restriction.
    
    **It can only be used if the mhn package was compiled with CUDA.**

    :param theta: matrix containing the theta entries of the current MHN
    :param mutation_data: StateContainer containing the mutation data the MHN should be trained on
    :return: tuple containing the normalized gradient and score
    
    :raise RuntimeError: this function raises a RuntimeError if the mhn package was not compiled with CUDA
    """
    IF NVCC_AVAILABLE:
        cdef int n = theta.shape[0]
        cdef int data_size = mutation_data.data_size

        cdef double score
        cdef np.ndarray[np.double_t] grad_out = np.empty(n * n, dtype=np.double)
        cdef int error_code
        cdef const char *error_name
        cdef const char *error_description

        error_code = cuda_gradient_and_score_implementation(&theta[0, 0], n, &mutation_data.states[0], data_size, &grad_out[0], &score)

        if error_code != 0:
            get_error_name_and_description(error_code, &error_name, &error_description)
            raise CUDAError(f'{error_name.decode("UTF-8")}: "{error_description.decode("UTF-8")}"')

        return (grad_out.reshape((n, n)) / data_size), (score / data_size)
    ELSE:
        raise RuntimeError('The mhn package was not compiled with CUDA, so you cannot use this function')


cpdef gradient_and_score(double[:, :] theta, StateContainer mutation_data):
    """
    This function computes the gradient using Cython AND CUDA (only if CUDA is installed)
    It will compute the gradients for data points with few mutations using the Cython implementation
    and compute the gradients for data points with many mutations using CUDA.
    If CUDA is not installed on your device, this function will only use the Cython implementation.
    
    :param theta: matrix containing the theta entries of the current MHN
    :param mutation_data: StateContainer containing the mutation data the MHN should be trained on
    :return: tuple containing the normalized gradient and score
    """
    IF NVCC_AVAILABLE:
        # number of mutations for which it is still faster to do it in cython
        # than using the cuda implementation
        # may vary depending on your device
        DEF critical_size = 13

        if mutation_data.get_max_mutation_num() <= critical_size:
            return cython_gradient_and_score(theta, mutation_data)

        cdef int data_size = mutation_data.data_size
        cdef int n = theta.shape[0]
        cdef State *sorted_data = <State*> malloc(data_size * sizeof(State));

        cdef int error_code
        cdef const char *error_name
        cdef const char *error_description

        cdef double *grad_out

        cdef int index_left, index_right
        cdef int i, j
        cdef State *state
        cdef int mutation_num

        cdef double score = 0

        index_left = 0
        index_right = data_size - 1

        # sort the data into samples that have more, and samples that have less
        # than *critical_size* mutations
        for i in range(data_size):
            mutation_num = get_mutation_num(&mutation_data.states[i])
            if mutation_num > critical_size:
                sorted_data[index_right] = mutation_data.states[i]
                index_right -= 1
            else:
                sorted_data[index_left] = mutation_data.states[i]
                index_left += 1

        cdef np.ndarray[np.double_t, ndim=2]  final_gradient = np.zeros((n, n), dtype=np.double)

        # check if there is any data point with more than *critical_size* mutations
        # and only call the CUDA function if this is the case
        if index_right != data_size - 1:
            grad_out = <double *> malloc(n*n * sizeof(double))
            error_code = cuda_gradient_and_score_implementation(&theta[0, 0], n, sorted_data + index_right + 1,
                                                                data_size - index_left, &grad_out[0], &score)

            if error_code != 0:
                get_error_name_and_description(error_code, &error_name, &error_description)
                raise CUDAError(f'{error_name.decode("UTF-8")}: "{error_description.decode("UTF-8")}"')

            for i in range(n):
                for j in range(n):
                    final_gradient[i, j] += grad_out[i*n + j]

            free(grad_out)

        cdef np.ndarray[np.double_t, ndim=2]  tmp_gradient = np.zeros((n, n), dtype=np.double)

        for i in range(index_left):
            state = &sorted_data[i]
            score += restricted_gradient_and_score(theta, state, tmp_gradient)
            final_gradient += tmp_gradient

        free(sorted_data)
        return final_gradient/data_size, score/data_size

    ELSE:
        return cython_gradient_and_score(theta, mutation_data)


CUDA_AVAILABLE = "CUDA is available"
CUDA_NOT_AVAILABLE = "CUDA compiler nvcc was not present during installation"
CUDA_NOT_FUNCTIONAL = "CUDA compiler nvcc available but CUDA functions not working. Check CUDA installation"
def cuda_available():
    """
    Call this function if you want to know if the mhn package is able to use CUDA functions on your device.
    """
    IF NVCC_AVAILABLE:
        if cuda_functional():
            return CUDA_AVAILABLE
        else:
            return CUDA_NOT_FUNCTIONAL
    ELSE:
        return CUDA_NOT_AVAILABLE




# by Stefan Vocht
#
# this file acts like a C header file for ModelConstruction.pyx
#
cimport numpy as np
np.import_array()

cpdef np.ndarray[np.double_t, ndim=1] q_subdiag(double[:, :] theta, int i)


cpdef np.ndarray[np.double_t, ndim=1] q_diag(double[:, :] theta)


"""
This submodule implements Likelihood.R from the original implementation in Cython.

It contains functions to compute the log-likelihood score and its gradient without state-space restriction as
well as functions for matrix-vector multiplications with the transition rate matrix and [I-Q]^(-1).
There are also functions to compute the probability distribution represented by an MHN and to sample artificial data
from a given MHN.
"""
# author(s): Stefan Vocht

cimport cython

from libc.stdlib cimport malloc, free

from .ModelConstruction cimport q_diag
from .PerformanceCriticalCode cimport internal_kron_vec, loop_j, compute_inverse

from scipy.linalg.cython_blas cimport dcopy, dscal, daxpy, ddot

import numpy as np
cimport numpy as np

np.import_array()

cdef internal_q_vec(double[:, :] theta, double[:] x, double[:] yout, bint diag = False, bint transp = False):
    """
    Multiplies the vector x with the matrix Q

    :param theta: thetas used to construct Q
    :param x: vector that is multiplied with Q
    :param yout: vector that contains the result of the multiplication
    :param diag: if False, the diagonal of Q is set to zero
    :param transp: if True, x is multiplied with Q^T

    :return: product of Q and x
    """
    cdef int n = theta.shape[0]
    cdef int nx = x.shape[0]
    cdef int i
    cdef double one_d = 1
    cdef int one_i = 1
    cdef double zero = 0

    cdef double *result_vec = <double *> malloc(sizeof(double) * nx)

    # initialize yout with zero
    dscal(&nx, &zero, &yout[0], &one_i)

    for i in range(n):
        internal_kron_vec(theta, i, x, result_vec, diag, transp)
        # add result of restricted_kronvec to yout
        daxpy(&nx, &one_d, result_vec, &one_i, &yout[0], &one_i)

    free(result_vec)


cpdef np.ndarray[np.double_t] q_vec(double[:, :] theta, double[:] x, bint diag = False, bint transp = False):
    """
    Multiplies the vector x with the matrix Q

    :param theta: thetas used to construct Q
    :param x: vector that is multiplied with Q
    :param diag: if False, the diagonal of Q is set to zero
    :param transp: if True, x is multiplied with Q^T

    :return: product of Q and x
    """
    cdef int n = theta.shape[0]
    cdef np.ndarray[np.double_t] result = np.empty(2**n, dtype=np.double)
    internal_q_vec(theta, x, result, diag, transp)
    return result


cpdef np.ndarray[np.double_t] jacobi(double[:, :] theta, np.ndarray[np.double_t] b, bint transp = False):
    """
    Returns the solution for [I - Q] x = b

    :param theta: thetas used to construct Q
    :param b:
    :param transp: if True, returns solution for ([I - Q]^-1)^T x = b
    :return:
    """
    cdef int n = theta.shape[1]
    cdef int i

    cdef np.ndarray[np.double_t] x = np.full(2**n, 1. / (2**n), dtype=np.double)
    cdef np.ndarray[np.double_t] dg = 1 - q_diag(theta)

    for i in range(n+1):
        x = b + q_vec(theta, x, diag=False, transp=transp)
        x = x / dg

    return x


cpdef np.ndarray[np.double_t] generate_pTh(double[:, :] theta, p0 = None):
    """
    Returns the probability distribution given by theta

    :param theta:
    :param p0:
    :return:
    """
    cdef int n = theta.shape[1]
    cdef np.ndarray[np.double_t] dg = 1 - q_diag(theta)
    cdef np.ndarray[np.double_t] pth = np.empty(2**n, dtype=np.double)

    if p0 is None:
        p0 = np.zeros(2**n)
        p0[0] = 1

    compute_inverse(theta, dg, p0, pth, False)
    return pth


def score(double[:, :] theta, np.ndarray[np.double_t] pD, np.ndarray[np.double_t] pth_space = None) -> float:
    """
    Calculates the score for the current theta

    :param theta:
    :param pD: probability distribution in the data
    :param pth_space: opional, with this parameter we can communicate with the function grad and use pth there again -> performance boost
    :return: score value
    """
    cdef np.ndarray[np.double_t] pth = generate_pTh(theta)

    if pth_space is not None:
        pth_space[:] = pth

    return pD.dot(np.log(pth))


def grad(double[:, :] theta, np.ndarray[np.double_t] pD, np.ndarray[np.double_t] pth_space = None) -> np.ndarray:
    """
    Implements gradient calculation of equation 7

    :param theta:
    :param pD: probability distribution of the training data
    :param pth: as pth is calculated in the score function anyways, we do not need to calculate it again
    :param pth_space: opional, with this parameter we can communicate with the function score and use pth here again -> performance boost
    :return: gradient you get from equation 7
    """
    cdef int n = theta.shape[0]
    cdef int nx = 1 << n
    cdef int i, j

    cdef np.ndarray[np.double_t] p0, pth

    cdef np.ndarray[np.double_t] dg = 1 - q_diag(theta)

    # distribution you get from our current model Theta (pth ~ "p_theta")
    if pth_space is None:
        # start distribution p_0 where no gene is mutated yet
        p0 = np.zeros(2 ** n)
        p0[0] = 1
        pth = np.empty(2**n, dtype=np.double)
        compute_inverse(theta, dg, p0, pth, False)

    else:
        pth = pth_space

    # should be (pD / pth)^T * R_theta^-1 from equation 7
    cdef np.ndarray[np.double_t] q = np.empty(2**n, dtype=np.double)
    compute_inverse(theta, dg, pD / pth, q, True)

    cdef np.ndarray[np.double_t, ndim=2] g = np.zeros((n, n))
    cdef double *r_vec = <double *> malloc(nx * sizeof(double))

    for i in range(n):
        internal_kron_vec(theta, i, pth, r_vec, diag=True, transp=False)
        for j in range(nx):
            r_vec[j] *= q[j]
        loop_j(i, n, r_vec, &g[0, 0])

    free(<void *> r_vec)
    return g


@cython.wraparound(False)
@cython.boundscheck(False)
def sample_artificial_data(np.ndarray[np.double_t, ndim=2] theta, int sample_num) -> np.ndarray:
    """
    Returns artificial data sampled from an MHN. Random values are generated with numpy, use np.random.seed()
    to make results reproducible.

    :param theta: theta matrix representing an MHN
    :param sample_num: number of samples in the generated data
    :returns: 2d numpy array in which every row corresponds to a sample and each column to a gene
    """
    cdef int n = theta.shape[0]
    cdef np.ndarray[np.double_t, ndim=2] exp_theta = np.exp(theta)
    cdef np.ndarray[np.int32_t, ndim=2] art_data = np.zeros((sample_num, n), dtype=np.int32)

    cdef np.ndarray[np.int32_t] in_current_sample = np.zeros(n, dtype=np.int32)
    cdef np.ndarray[np.int32_t] possible_gene_mutations = np.empty(n, dtype=np.int32)
    cdef np.ndarray[np.double_t] rates_from_current_state = np.empty(n, dtype=np.double)

    cdef int j, gene, mutated_gene
    cdef int sample_index, current_sample_length
    cdef double observation_time, current_time, passed_time
    cdef double sum_rates, rate, random_crit, accumulated_rate

    for sample_index in range(sample_num):
        in_current_sample[:] = 0
        current_sample_length = 0
        observation_time = np.random.exponential(1)
        current_time = 0.
        while current_sample_length < n:
            sum_rates = 0.
            j = 0
            for gene in range(n):
                if not in_current_sample[gene]:
                    possible_gene_mutations[j] = gene
                    rate = exp_theta[gene, gene]
                    for mutated_gene in range(n):
                        if in_current_sample[mutated_gene]:
                            rate *= exp_theta[gene, mutated_gene]
                    rates_from_current_state[j] = rate
                    sum_rates += rate
                    j += 1

            passed_time = np.random.exponential(1/sum_rates)
            current_time += passed_time
            if current_time > observation_time:
                break
            random_crit = np.random.random(1)[0] * sum_rates
            accumulated_rate = 0.
            for j in range(n - current_sample_length):
                rate = rates_from_current_state[j]
                gene = possible_gene_mutations[j]
                accumulated_rate += rate
                if random_crit <= accumulated_rate:
                    in_current_sample[gene] = 1
                    current_sample_length += 1
                    art_data[sample_index, gene] = 1
                    break
    return art_data


IF NVCC_AVAILABLE:
    cdef extern from *:
        """
        #ifdef _WIN32
        #define DLL_PREFIX __declspec(dllexport)
        #else
        #define DLL_PREFIX
        #endif

        int DLL_PREFIX cuda_full_state_space_gradient_score(double *ptheta, int n, double *pD, double *grad_out, double *score_out);
        void DLL_PREFIX gpu_compute_inverse(double *theta, int n, double *b, double *xout, int transp);
        """
        int cuda_full_state_space_gradient_score(double *ptheta, int n, double *pD, double *grad_out, double *score_out)
        void gpu_compute_inverse(double *theta, int n, double *b, double *xout, int transp)


def cuda_gradient_and_score(double[:, :] theta, double[:] pD):
    """
    Computes the log-likelihood score and its gradient for a given theta and a given empirical distribution.

    **It can only be used if the mhn package was compiled with CUDA.**

    :param theta: theta matrix representing the MHN
    :param pD: probability distribution according to the training data
    :returns: tuple containing the gradient and the score

    :raise RuntimeError: this function raises a RuntimeError if the mhn package was not compiled with CUDA
    """

    IF NVCC_AVAILABLE:
        cdef int n = theta.shape[0]
        cdef double score
        cdef np.ndarray[np.double_t, ndim=2] gradient = np.empty((n, n), dtype=np.double)
        cdef int error_code

        error_code = cuda_full_state_space_gradient_score(&theta[0, 0], n, &pD[0], &gradient[0, 0], &score)

        return gradient, score
    ELSE:
        raise RuntimeError('The mhn package was not compiled with CUDA, so you cannot use this function.')


def cuda_compute_inverse(double[:, :] theta, double[:] b, bint transp = False):
    """
    Computes the solution for [I-Q] x = b on the GPU.

    **It can only be used if the mhn package was compiled with CUDA.**

    :param theta: theta matrix representing the MHN
    :param b: vector that is multiplied with [I-Q]^(-1)
    :param transp: if set to True, computes solution for [I-Q]^T x = b
    :returns: the solution for [I-Q] x = b

    :raise RuntimeError: this function raises a RuntimeError if the mhn package was not compiled with CUDA
    """

    IF NVCC_AVAILABLE:
        cdef int n = theta.shape[0]
        cdef int nx = 1 << n
        cdef np.ndarray[np.double_t] xout = np.empty(nx, dtype=np.double)
        gpu_compute_inverse(&theta[0, 0], n, &b[0], &xout[0], transp)
        return xout
    ELSE:
        raise RuntimeError('The mhn package was not compiled with CUDA, so you cannot use this function.')


import unittest

from mhn.original import ModelConstruction
from mhn.ssr import matrix_exponential, state_containers

import numpy as np
import scipy


class TestMatrixExponential(unittest.TestCase):

    def setUp(self) -> None:
        np.random.seed(0)

    def test_correctness_expm(self):
        n = 6
        theta = ModelConstruction.random_theta(n)
        q = ModelConstruction.build_q(theta)
        all_mutated_data = np.ones((1, n), dtype=np.int32)
        b = np.random.random(2**n)
        eps = 1e-4

        for t in np.arange(0, 6, 0.6):
            container = state_containers.StateAgeContainer(all_mutated_data, np.array([t]))
            result1 = scipy.linalg.expm(t * q).dot(b)
            result2 = matrix_exponential.restricted_expm(theta, b, container, eps)
            self.assertTrue(np.abs(result2 - result1).sum() < eps)

    def test_calc_gamma_numerically(self):
        n = 6
        theta = ModelConstruction.random_theta(n)
        np_data_matrix = np.zeros((1, 6), dtype=np.int32)
        np_data_matrix[:, 2:4] = 1
        container = state_containers.StateAgeContainer(np_data_matrix, np.array([0.5]))
        h = 1e-10
        for i in range(n):
            for j in range(n):
                gamma1, dgamma = matrix_exponential.calc_gamma_wrapper(theta, container, i, j)
                theta_copy = theta.copy()
                theta_copy[i, j] += h
                gamma2, _ = matrix_exponential.calc_gamma_wrapper(theta_copy, container, i, j)
                numerical_deriv = (gamma2 - gamma1) / h
                self.assertAlmostEqual(numerical_deriv, dgamma, 3)

    def test_score_and_gradient_numerically(self):
        n = 6
        sample_num = 5
        assert sample_num < n, "This test needs the sample number to be smaller than n to work because of how the " \
                               "data is constructed "
        theta = ModelConstruction.random_theta(n)
        np_data_matrix = np.zeros((sample_num, n), dtype=np.int32)
        for i in range(sample_num):
            np_data_matrix[i, :i] = 1

        ages = np.linspace(0, 6, sample_num)
        states_and_ages = state_containers.StateAgeContainer(np_data_matrix, ages)
        h = 1e-10

        gradient, old_score = matrix_exponential.cython_gradient_and_score(theta, states_and_ages, 1e-6)
        for i in range(n):
            for j in range(n):
                theta_copy = theta.copy()
                theta_copy[i, j] += h
                _, new_score = matrix_exponential.cython_gradient_and_score(theta_copy, states_and_ages, 1e-6)
                numerical_gradient = (new_score - old_score) / h
                self.assertAlmostEqual(numerical_gradient, gradient[i, j], 3)

    def test_sort_by_age(self):
        """
        StateAgeContainer should automatically sort the samples according to their age.
        """
        n = 6
        sample_num = 5
        assert sample_num < n, "This test needs the sample number to be smaller than n to work because of how the " \
                               "data is constructed "
        theta = ModelConstruction.random_theta(n)
        np_data_matrix = np.zeros((sample_num, n), dtype=np.int32)
        for i in range(sample_num):
            np_data_matrix[i, :i] = 1
        ages = np.linspace(0, 6, sample_num)

        states_and_ages = state_containers.StateAgeContainer(np_data_matrix, ages)
        grad1, score1 = matrix_exponential.cython_gradient_and_score(theta, states_and_ages, 1e-6)

        permutation = np.random.permutation(sample_num)
        np_data_matrix = np_data_matrix[permutation]
        ages = ages[permutation]
        states_and_ages = state_containers.StateAgeContainer(np_data_matrix, ages)
        grad2, score2 = matrix_exponential.cython_gradient_and_score(theta, states_and_ages, 1e-6)

        self.assertAlmostEqual(score1, score2, 8)
        np.testing.assert_array_equal(np.around(grad1, decimals=5), np.around(grad2, decimals=5))


class TestCudaMatrixExponential(unittest.TestCase):

    def setUp(self) -> None:
        np.random.seed(0)
        self.skipTest("There is currently no CUDA implementation for the matrix exponential available. "
                      "Might be added in the future.")

    def test_compare_with_cython(self):
        n = 6
        sample_num = 5
        assert sample_num < n, "This test needs the sample number to be smaller than n to work because of how the " \
                               "data is constructed "
        theta = ModelConstruction.random_theta(n)
        np_data_matrix = np.zeros((sample_num, n), dtype=np.int32)
        for i in range(sample_num):
            np_data_matrix[i, :i+1] = 1

        ages = np.linspace(0, 6, sample_num)
        states_and_ages = state_containers.StateAgeContainer(np_data_matrix, ages)

        grad1, score1 = matrix_exponential.cython_gradient_and_score(theta, states_and_ages, 1e-6)
        grad2, score2 = matrix_exponential.cuda_gradient_and_score(theta, states_and_ages, 1e-6)

        self.assertAlmostEqual(np.max(np.abs(grad2 - grad1)), 0, 5)
        self.assertAlmostEqual(score1, score2, 5)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
import numpy

import sys
import os
import platform
from shutil import which
import subprocess

IS_WINDOWS = (platform.system() == 'Windows')   # get the operating system
STATE_SIZE = 8                                  # the compiled code supports MHNs with maximum size of 32 * STATE_SIZE
GENERATE_DEBUG_HTML = False                     # set this to True so that Cython generates a optimization HTML file

assert STATE_SIZE > 0                           # make sure STATE_SIZE is greater zero

with open("README.md", 'r') as f:
    long_description = f.read()


def compile_cuda_code(folder, cuda_filename, lib_name, *extra_compile_args, additional_cuda_files=None):
    """
    This function compiles the CUDA code of this package to run functions on the GPU
    """
    if additional_cuda_files is None:
        additional_cuda_files = []
    if IS_WINDOWS:
        output_filename = os.path.join(folder, f"{lib_name}.dll")
    else:
        output_filename = os.path.join(folder, f"lib{lib_name}.so")

    cuda_filename = os.path.join(folder, cuda_filename)
    # check if the shared library file was modified after the source files
    shared_lib_latest_version = True
    try:
        for cuda_file in [cuda_filename] + additional_cuda_files:
            shared_lib_latest_version &= (os.path.getmtime(cuda_file) - os.path.getmtime(output_filename) < 0)
    except FileNotFoundError:
        shared_lib_latest_version = False

    # check if compilation is even necessary or if the CUDA file was already compiled
    # for that we check if the shared library file exists and if it was modified after the source file
    # if the argument --force is given to setup.py, the CUDA code is always recompiled
    if "--force" not in sys.argv and os.path.isfile(output_filename) and shared_lib_latest_version:
        print("No need to compile CUDA code again...")
        return

    # command to compile the CUDA code using nvcc
    compile_command = ['nvcc', '-o', output_filename, '--shared', cuda_filename, *additional_cuda_files,
                       f'-DSTATE_SIZE={STATE_SIZE}', *extra_compile_args]
    if not IS_WINDOWS:
        compile_command += ['-Xcompiler', '-fPIC']

    # execute command and print the output
    print(subprocess.run(compile_command, stdout=subprocess.PIPE).stdout.decode('utf-8', "ignore"))


# check if nvcc (the cuda compiler) is available on the device
nvcc_available = int(which('nvcc') is not None)

libraries = []
# only compile CUDA code if nvcc is available and if we do not create a source distribution
if nvcc_available and 'sdist' not in sys.argv:
    libraries.append("CudaStateSpaceRestriction")
    libraries.append("CudaFullStateSpace")
    compile_cuda_code("./mhn/original/", "cuda_full_state_space.cu", "CudaFullStateSpace",
                      additional_cuda_files=["./mhn/original/cuda_inverse_by_substitution.cu"])
    compile_cuda_code("./mhn/ssr/", "cuda_state_space_restriction.cu", "CudaStateSpaceRestriction", f'-I./mhn/original/',
                      additional_cuda_files=["./mhn/original/cuda_inverse_by_substitution.cu"])


# define compile options for the Cython files
ext_modules = [
    Extension(
        "mhn.ssr.state_containers",
        ["./mhn/ssr/state_containers.pyx"],
        extra_compile_args=[
            f'-DSTATE_SIZE={STATE_SIZE}'
        ]
    ),
    Extension(
        "mhn.ssr.state_space_restriction",
        ["./mhn/ssr/state_space_restriction.pyx"],
        libraries=libraries,
        library_dirs=["./mhn/ssr/", "./mhn/original/"],
        include_dirs=['./mhn/ssr/', "./mhn/original/"],
        extra_compile_args=[
            '/Ox' if IS_WINDOWS else '-O2',
            f'-DSTATE_SIZE={STATE_SIZE}'
        ],
        extra_link_args=[] if IS_WINDOWS else [
            '-Wl,-rpath,$ORIGIN',
            '-Wl,-rpath-link,../original/',
        ]
    ),
    Extension(
        "mhn.ssr.matrix_exponential",
        ["./mhn/ssr/matrix_exponential.pyx"],
        libraries=libraries,
        library_dirs=["./mhn/ssr/", "./mhn/original/"],
        # runtime_library_dirs=None if IS_WINDOWS else ["./mhn/ssr/", "./mhn/original/"],
        include_dirs=['./mhn/ssr/', "./mhn/original/"],
        extra_compile_args=[
            '/Ox' if IS_WINDOWS else '-O2',
            f'-DSTATE_SIZE={STATE_SIZE}'
        ],
        extra_link_args=[]
    ),
    Extension(
        "mhn.original.Likelihood",
        ["./mhn/original/Likelihood.pyx"],
        libraries=libraries,
        library_dirs=["./mhn/ssr/", "./mhn/original/"],
        include_dirs=['./mhn/ssr/', "./mhn/original/"],
        extra_compile_args=[
            '/Ox' if IS_WINDOWS else '-O2'
        ],
        extra_link_args=[] if IS_WINDOWS else [
            '-Wl,-rpath,$ORIGIN',
            '-Wl,-rpath-link,../ssr/',
        ]
    ),
    Extension(
        "mhn.original.PerformanceCriticalCode",
        ["./mhn/original/PerformanceCriticalCode.pyx"],
        extra_compile_args=[
            '/Ox' if IS_WINDOWS else '-O2',
        ]
    ),
    Extension(
        "mhn.original.ModelConstruction",
        ["./mhn/original/ModelConstruction.pyx"],
        extra_compile_args=[
            '/Ox' if IS_WINDOWS else '-O2'
        ]
    )
]

# we only want the source code in a source distribution
if 'sdist' in sys.argv:
    ext_modules = []

setup(
    name="mhn",
    version="0.0.13",
    packages=find_packages(),
    author="Stefan Vocht, Kevin Rupp, Y. Linda Hu",
    description="A package to train and work with Mutual Hazard Networks",
    long_description=long_description,
    long_description_content_type='text/markdown',
    ext_modules=cythonize(ext_modules,
                          annotate=GENERATE_DEBUG_HTML,
                          compile_time_env=dict(
                                                NVCC_AVAILABLE=nvcc_available,
                                                STATE_SIZE=STATE_SIZE
                                                ),
                          compiler_directives={'embedsignature': True}
                          ),
    include_dirs=[numpy.get_include()],
    include_package_data=True,
    install_requires=[
        'numpy>=1.23.0',
        'scipy>=1.8.0',
        'pandas>=1.5.3'
    ],
    python_requires='>=3.8'
)



"""
This submodule implements RegularizedOptimization.R in Python.

It contains functions to learn an MHN on the *full state-space* for a given data distribution and implements the L1 regularization.
"""
# author(s): Stefan Vocht

import numpy as np
from scipy.optimize import minimize

from . import Likelihood
from . import ModelConstruction

from typing import Callable


def L1(theta: np.ndarray, eps: float = 1e-05) -> float:
    """
    Computes the L1 penalty

    :param theta: the theta matrix representing the MHN
    :param eps: small epsilon value, mainly there for the derivative

    :returns: the L1 penalty for the given theta matrix
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return np.sum(np.sqrt(theta_**2 + eps))


def L1_(theta: np.ndarray, eps: float = 1e-05) -> np.ndarray:
    """
    Derivative of the L1 penalty

    :param theta: the theta matrix representing the MHN
    :param eps: small epsilon value that makes sure that we don't divide by zero

    :returns: the derivative of the L1 penalty
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return theta_ / np.sqrt(theta_**2 + eps)


def score_reg(theta: np.ndarray, pD: np.ndarray, lam: float, n: int = None, pth_space: np.ndarray = None) -> float:
    """
    Score with L1 - regularization

    :param theta: the theta matrix representing the MHN
    :param pD: distribution given by the training data
    :param lam: tuning parameter lambda for regularization
    :param n: number of columns/rows of theta
    :param pth_space: optional, with this parameter we can communicate with the gradient function and use pth there again -> performance boost

    :returns: the score of the current MHN penalized with the L1 regularization
    """
    n = n or int(np.sqrt(theta.size))
    theta = theta.reshape((n, n))

    return -(Likelihood.score(theta, pD, pth_space) - lam * L1(theta))


def grad_reg(theta: np.ndarray, pD: np.ndarray, lam: float, n: int = 0, pth_space: np.ndarray = None) -> np.ndarray:
    """
    Gradient with L1 - regularization

    :param theta: the theta matrix representing the MHN
    :param pD: distribution given by the training data
    :param lam: tuning parameter lambda for regularization
    :param n: number of columns/rows of theta
    :param pth_space: optional, as pth is calculated in the score function anyway, we do not need to calculate it again -> performance boost

    :return: the gradient of the L1 - regularized score
    """
    n = n or int(np.sqrt(theta.size))
    theta_ = theta.reshape((n, n))

    return -(Likelihood.grad(theta_, pD, pth_space) - lam * L1_(theta_)).flatten()


def learn_MHN(pD: np.ndarray, init: np.ndarray = None, lam: float = 0, maxit: int = 5000,
              trace: bool = False, reltol: float = 1e-07, round_result: bool = True,
              callback: Callable = None, score_func: Callable = score_reg, jacobi: Callable = grad_reg) -> np.ndarray:
    """
    This function is used to train an MHN to a given probability distribution pD.

    :param pD: probability distribution used to train the new model
    :param init: starting point for the training (initial theta)
    :param lam: tuning parameter lambda for regularization
    :param maxit: maximum number of training iterations
    :param trace: set to True to print convergence messages (see scipy.optimize.minimize)
    :param reltol: Gradient norm must be less than reltol before successful termination (see "gtol" scipy.optimize.minimize)
    :param round_result: if True, the result is rounded to two decimal places
    :param callback: function called after each iteration, must take theta as argument
    :param score_func: score function used for training
    :param jacobi: gradient function used for training
    :return: trained model
    """

    n = int(np.log2(pD.size))

    if init is None:
        init = ModelConstruction.learn_indep(pD)

    pth_space = Likelihood.generate_pTh(init)

    opt = minimize(fun=score_func, x0=init, args=(pD, lam, n, pth_space), method="BFGS", jac=jacobi,
                   options={'maxiter': maxit, 'disp': trace, 'gtol': reltol}, callback=callback)

    theta = opt.x.reshape((n, n))

    if round_result:
        theta = np.around(theta, decimals=2)

    return theta



"""
This submodule contains a class to represent Mutual Hazard Networks
"""
# author(s): Y. Linda Hu, Stefan Vocht

from __future__ import annotations

from .original import Likelihood
from .ssr import state_space_restriction

import numpy as np
import pandas as pd
import json


class MHN:
    """
    This class represents the Mutual Hazard Network.
    """

    def __init__(self, log_theta: np.array, events: list[str] = None, meta: dict = None):
        """
        :param log_theta: logarithmic values of the theta matrix representing the MHN
        :param events: (optional) list of strings containing the names of the events considered by the MHN
        :param meta: (optional) dictionary containing metadata for the MHN, e.g. parameters used to train the model
        """

        self.log_theta = log_theta
        self.events = events
        self.meta = meta

    def sample_artificial_data(self, sample_num: int, as_dataframe: bool = False) -> np.ndarray | pd.DataFrame:
        """
        Returns artificial data sampled from this MHN. Random values are generated with numpy, use np.random.seed()
        to make results reproducible.

        :param sample_num: number of samples in the generated data
        :param as_dataframe: if True, the data is returned as a pandas DataFrame, else numpy matrix

        :returns: array or DataFrame with samples as rows and events as columns
        """
        art_data = Likelihood.sample_artificial_data(self.log_theta, sample_num)
        if as_dataframe:
            df = pd.DataFrame(art_data)
            if self.events is not None:
                df.columns = self.events
            return df
        else:
            return art_data

    def compute_marginal_likelihood(self, state: np.ndarray) -> float:
        """
        Computes the likelihood of observing a given state, where we consider the observation time to be an
        exponential random variable with mean 1.

        :param state: a 1d numpy array (dtype=np.int32) containing 0s and 1s, where each entry represents an event being present (1) or not (0)

        :returns: the likelihood of observing the given state according to this MHN
        """
        if not set(state.flatten()).issubset({0, 1}):
            raise ValueError("The state array must only contain 0s and 1s")
        mutation_num = np.sum(state)
        nx = 1 << mutation_num
        p0 = np.zeros(nx)
        p0[0] = 1
        p_th = state_space_restriction.compute_restricted_inverse(self.log_theta, state, p0, False)
        return p_th[-1]

    def save(self, filename: str):
        """
        Save the MHN in a CSV file. If metadata is given, it will be stored in a separate JSON file.

        :param filename: name of the CSV file without(!) the '.csv', JSON will be named accordingly
        """
        pd.DataFrame(self.log_theta, columns=self.events,
                     index=self.events).to_csv(f"{filename}.csv")
        if self.meta is not None:
            with open(f"{filename}_meta.json", "x") as file:
                json.dump(self.meta, file, indent=4)

    @classmethod
    def load(cls, filename: str, events: list[str] = None) -> MHN:
        """
        Load an MHN object from a CSV file.

        :param filename: name of the CSV file without(!) the '.csv'
        :param events: list of strings containing the names of the events considered by the MHN

        :returns: MHN object
        """
        df = pd.read_csv(f"{filename}.csv", index_col=0)
        if events is None and (df.columns != pd.Index([str(x) for x in range(len(df.columns))])).any():
            events = df.columns.to_list()
        try:
            with open(f"{filename}_meta.json", "r") as file:
                meta = json.load(file)
        except FileNotFoundError:
            meta = None
        return cls(np.array(df), events=events, meta=meta)



# by Stefan Vocht
#
# this file contains unittests for state_space_restriction.pyx
#

import unittest
import numpy as np
from mhn.ssr import state_space_restriction
from mhn.ssr.state_containers import StateContainer
from mhn.original import Likelihood, UtilityFunctions, ModelConstruction


class TestCythonGradient(unittest.TestCase):
    """
    Tests for the function cython_gradient_and_score
    """
    def setUp(self) -> None:
        """
        Preparation for each test
        """
        np.random.seed(0)  # set random seed for reproducibility

    def test_comparison_with_numerical_gradient(self):
        """
        Computes the gradient numerically and compares it with the analytical one
        """
        n = 5
        sample_num = 30
        h = 1e-10
        theta = ModelConstruction.random_theta(n)
        random_sample = np.random.choice([0, 1], (sample_num, n), p=[0.5, 0.5]).astype(np.int32)
        pD = UtilityFunctions.data_to_pD(random_sample)                             # compute the data distribution
        original_score = Likelihood.score(theta, pD)
        # compute the gradient numerically
        numerical_gradient = np.empty((n, n), dtype=float)
        for i in range(n):
            for j in range(n):
                theta_copy = theta.copy()
                theta_copy[i, j] += h
                new_score = Likelihood.score(theta_copy, pD)
                numerical_gradient[i, j] = (new_score - original_score) / h

        analytic_gradient, score = state_space_restriction.cython_gradient_and_score(theta, StateContainer(random_sample))
        self.assertEqual(round(score, 8), round(original_score, 8))
        np.testing.assert_array_equal(np.around(numerical_gradient, decimals=3), np.around(analytic_gradient, decimals=3))

    def test_gene_position_permutation(self):
        """
        Permutation of the position of genes in the data should lead to a permutation in the gradient,
        but should not change the score
        """
        n = 40  # make n > 32 to make sure that the logic of the "State" struct in the C implementation works as expected
        sample_num = 30
        theta = ModelConstruction.random_theta(n)
        random_sample = np.random.choice([0, 1], (sample_num, n), p=[0.8, 0.2]).astype(np.int32)
        # make sure that there are mutations in two different "parts" of the "State" C struct
        random_sample[:, 0] = 1
        random_sample[:, -1] = 1
        # compute original gradient and score
        gradient1, score1 = state_space_restriction.cython_gradient_and_score(theta, StateContainer(random_sample))
        # permute the sample and theta, compute gradient and score and reverse the permutation
        permutation = np.random.permutation(n)
        reverse = np.empty(n, int)
        reverse[permutation] = np.arange(n)
        permutation_sample = random_sample[:, permutation]
        permutation_theta = theta[permutation][:, permutation]
        gradient2, score2 = state_space_restriction.cython_gradient_and_score(permutation_theta, StateContainer(permutation_sample))
        reversed_gradient = gradient2[reverse][:, reverse]
        np.testing.assert_array_equal(permutation_sample[:, reverse], random_sample)
        # compare gradients and scores
        np.testing.assert_array_equal(np.around(gradient1, decimals=8), np.around(reversed_gradient, decimals=8))
        self.assertEqual(score1, score2)


class TestCudaGradient(unittest.TestCase):
    """
    Tests for the function cuda_gradient_and_score
    """
    def setUp(self) -> None:
        """
        Preparation for each test
        """
        np.random.seed(0)  # set random seed for reproducibility
        if state_space_restriction.cuda_available() != state_space_restriction.CUDA_AVAILABLE:
            self.skipTest("CUDA not available for testing")

    def test_compare_with_cython(self):
        n = 40
        sample_num = 30
        theta = ModelConstruction.random_theta(n)
        random_sample = np.random.choice([0, 1], (sample_num, n), p=[0.7, 0.3]).astype(np.int32)
        # make sure that there are mutations in two different "parts" of the "State" C struct
        random_sample[:, 1] = 1
        random_sample[:, -3] = 1
        state_containers = StateContainer(random_sample)
        gradient1, score1 = state_space_restriction.cython_gradient_and_score(theta, state_containers)
        gradient2, score2 = state_space_restriction.cuda_gradient_and_score(theta, state_containers)
        self.assertEqual(round(score1, 8), round(score2, 8))
        np.testing.assert_array_equal(np.around(gradient1, decimals=8), np.around(gradient2, decimals=8))

    def test_gene_position_permutation(self):
        """
        Permutation of the position of genes in the data should lead to a permutation in the gradient,
        but should not change the score
        """
        n = 40  # make n > 32 to make sure that the logic of the "State" struct in the C implementation works as expected
        sample_num = 30
        theta = ModelConstruction.random_theta(n)
        random_sample = np.random.choice([0, 1], (sample_num, n), p=[0.8, 0.2]).astype(np.int32)
        # make sure that there are mutations in two different "parts" of the "State" C struct
        random_sample[:, 1] = 1
        random_sample[:, -3] = 1
        # compute original gradient and score
        gradient1, score1 = state_space_restriction.cuda_gradient_and_score(theta, StateContainer(random_sample))
        # permute the sample and theta, compute gradient and score and reverse the permutation
        permutation = np.random.permutation(n)
        reverse = np.empty(n, int)
        reverse[permutation] = np.arange(n)
        permutation_sample = random_sample[:, permutation]
        permutation_theta = theta[permutation][:, permutation]
        gradient2, score2 = state_space_restriction.cuda_gradient_and_score(permutation_theta.copy(), StateContainer(permutation_sample))
        reversed_gradient = gradient2[reverse][:, reverse]
        np.testing.assert_array_equal(permutation_sample[:, reverse], random_sample)
        # compare gradients and scores
        np.testing.assert_array_equal(np.around(gradient1, decimals=8), np.around(reversed_gradient, decimals=8))
        self.assertEqual(score1, score2)

    def test_auto_device_picker(self):
        """
        Tests the gradient_and_score() function, which uses both the CPU and the GPU.
        """
        n = 40
        sample_num = 30
        theta = ModelConstruction.random_theta(n)
        random_sample = np.random.choice([0, 1], (sample_num, n), p=[0.7, 0.3]).astype(np.int32)
        # make sure that there are mutations in two different "parts" of the "State" C struct
        random_sample[:, 1] = 1
        random_sample[:, -3] = 1
        # one sample with more than critical_size mutations -> GPU
        random_sample[0, :] = 0
        random_sample[0, :22] = 1
        # one sample with less than critical_size mutations -> CPU
        random_sample[1, :] = 0
        random_sample[1, :5] = 1
        state_containers = StateContainer(random_sample)
        gradient1, score1 = state_space_restriction.gradient_and_score(theta, state_containers)
        gradient2, score2 = state_space_restriction.cuda_gradient_and_score(theta, state_containers)
        self.assertEqual(round(score1, 8), round(score2, 8))
        np.testing.assert_array_equal(np.around(gradient1, decimals=8), np.around(gradient2, decimals=8))


if __name__ == '__main__':
    unittest.main()



"""
This part of the package contains functions that use state-space restriction (SSR) to become more efficient and
compute scores and their gradients that are needed for learning MHNs much faster.
"""

from . import state_containers
from . import state_space_restriction
from . import matrix_exponential
